'''Provides 'base' Nawah Core Module'''

from nawah.classes import Attr, Func, Module, Perm

base = Module(
    name='base',
    desc='\'base\' module provides necessary structure to allow modules to access Base '
    'Functions Callables via \'call\' Utility',
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            call_args={'raise_no_success': Attr.BOOL()},
        ),
        'create': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            call_args={'raise_no_success': Attr.BOOL()},
        ),
        'update': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            call_args={'raise_no_success': Attr.BOOL()},
        ),
        'delete': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            call_args={'raise_no_success': Attr.BOOL()},
        ),
        'obtain_lock': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            doc_attrs={'tags': Attr.LIST(list=[Attr.STR()]), 'attempts': Attr.INT()},
        ),
        'delete_lock': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            query_attrs={'_id': Attr.ID()},
        ),
    },
)



'''Provides Exception classes for Base Functions'''

# START: Shared exceptions
class UtilityModuleDataCallException(Exception):
    '''Raised if a data call was made for a Utility Module'''

    status = 400

    def __init__(self, *, module_name, func_name):
        super().__init__(
            UtilityModuleDataCallException.format_msg(
                module_name=module_name, func_name=func_name
            ),
            {'module_name': module_name, 'func_name': func_name},
        )

    @staticmethod
    def format_msg(*, module_name, func_name):
        '''Formats exception message'''

        return (
            f'Nawah Function \'{func_name}\' can\'t be called on Nawah Utility module '
            f'\'{module_name}\''
        )


class DuplicateUniqueException(Exception):
    '''Raised if 'create', 'update' call has at least one 'unique_attr' value in 'doc'
    that matches another doc'''

    status = 400

    def __init__(self, *, unique_attrs):
        super().__init__(
            DuplicateUniqueException.format_msg(unique_attrs=unique_attrs),
            {'unique_attrs': unique_attrs},
        )

    @staticmethod
    def format_msg(*, unique_attrs):
        '''Formats exception message'''

        return f'A doc matching at least one of \'{unique_attrs}\' already exists'


class InvalidDocException(Exception):
    '''Raised if 'delete_file', 'retrieve_file' call 'query' does not return any doc'''

    status = 400

    def __init__(self, *, doc_id):
        super().__init__(
            InvalidDocException.format_msg(doc_id=doc_id), {'doc_id': doc_id}
        )

    @staticmethod
    def format_msg(*, doc_id):
        '''Formats exception message'''

        return f'Query for doc with \'_id\' \'{doc_id}\' returned no results'


# END: Shared exceptions

# START: create exceptions
class NoDocCreatedException(Exception):
    '''Raised if 'create' call creates no doc'''

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocCreatedException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'No documents were created for module \'{module_name}\''


# END: create exceptions

# START: delete exceptions
class NoDocDeletedException(Exception):
    '''Raised if 'delete' call deletes no doc'''

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocDeletedException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'No documents were deleted for module \'{module_name}\''


# END: delete exceptions

# START: read exceptions
class NoDocFoundException(Exception):
    '''Raised if 'read' call founds no doc'''

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocFoundException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'No documents were found for module \'{module_name}\''


# END: read exceptions

# START: delete_file exceptions
class InvalidDeleteFileDocAttrException(Exception):
    '''Raised if 'delete_file' call 'query' returns doc with attr not valid for operation'''

    status = 400

    def __init__(self, *, doc_id, attr_name):
        super().__init__(
            InvalidDeleteFileDocAttrException.format_msg(
                doc_id=doc_id, attr_name=attr_name
            ),
            {'doc_id': doc_id, 'attr_name': attr_name},
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name):
        '''Formats exception message'''

        return f'Query for doc with \'_id\' \'{doc_id}\' returned doc with invalid value for attr \'{attr_name}\''


class InvalidDeleteFileIndexException(Exception):
    '''Raised if 'delete_file' call 'query' returns doc with index for call not valid for operation'''

    status = 400

    def __init__(self, *, doc_id, attr_name, index):
        super().__init__(
            InvalidDeleteFileIndexException.format_msg(
                doc_id=doc_id, attr_name=attr_name, index=index
            ),
            {'doc_id': doc_id, 'attr_name': attr_name, 'index': index},
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, index):
        '''Formats exception message'''

        return f'Index \'{index}\' for attr \'{attr_name}\' for doc with \'_id\' \'{doc_id}\' is invalid'


class InvalidDeleteFileIndexValueException(Exception):
    '''Raised if 'delete_file' call 'query' returns doc with index value for call not valid for operation'''

    status = 400

    def __init__(self, *, doc_id, attr_name, index, index_val_type):
        super().__init__(
            InvalidDeleteFileIndexValueException.format_msg(
                doc_id=doc_id,
                attr_name=attr_name,
                index=index,
                index_val_type=index_val_type,
            ),
            {
                'doc_id': doc_id,
                'attr_name': attr_name,
                'index': index,
                'index_val_type': index_val_type,
            },
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, index, index_val_type):
        '''Formats exception message'''

        return f'Index \'{index}\' for attr \'{attr_name}\' for doc with \'_id\' \'{doc_id}\' is of invalid type \'{index_val_type}\''


class InvalidDeleteFileMismatchNameException(Exception):
    '''Raised if 'delete_file' call 'query' returns doc with index value file name not matching 'query' '''

    status = 400

    def __init__(self, *, doc_id, attr_name, index, query_file_name, index_file_name):
        super().__init__(
            InvalidDeleteFileMismatchNameException.format_msg(
                doc_id=doc_id,
                attr_name=attr_name,
                index=index,
                query_file_name=query_file_name,
                index_file_name=index_file_name,
            ),
            {
                'doc_id': doc_id,
                'attr_name': attr_name,
                'index': index,
                'query_file_name': query_file_name,
                'index_file_name': index_file_name,
            },
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, index, query_file_name, index_file_name):
        '''Formats exception message'''

        return f'Index \'{index}\' for attr \'{attr_name}\' for doc with \'_id\' \'{doc_id}\' is for file \'{index_file_name}\', not \'{query_file_name}\''


# END: delete_file exceptions

# START: obtain_lock exceptions
class FailedObtainLockException(Exception):
    '''Raised if 'obtain_lock' call failed to obtain lock'''

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            FailedObtainLockException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'Failed to obtain lock for \'{module_name}\''


# END: obtain_lock exceptions

# START: delete_lock exceptions
class FailedDeleteLockException(Exception):
    '''Raised if 'obtain_lock','delete_lock' call failed to deleted obtaiend lock'''

    status = 500

    def __init__(self, *, module_name, lock_id):
        super().__init__(
            FailedDeleteLockException.format_msg(
                module_name=module_name, lock_id=lock_id
            ),
            {'module_name': module_name, 'lock_id': lock_id},
        )

    @staticmethod
    def format_msg(*, module_name, lock_id):
        '''Formats exception message'''

        return f'Failed to delete lock \'{lock_id}\' for \'{module_name}\'. Delete manually now'


# END: delete_lock exceptions

# START: update exceptions
class UpdateMultiUniqueException(Exception):
    '''Raised if 'update' call is attempted on multiple docs
    where at least one 'unique_attr' is present in 'update_doc' '''

    status = 400

    def __init__(self):
        super().__init__(UpdateMultiUniqueException.format_msg())

    @staticmethod
    def format_msg():
        '''Formats exception message'''

        return (
            'Update call query has more than one doc as results. This would result in'
            'duplication'
        )


class EmptyUpdateDocException(Exception):
    '''Raised if 'update' call has empty \'doc\' '''

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            EmptyUpdateDocException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'Update \'doc\' is empty for module \'{module_name}\''


class NoDocUpdatedException(Exception):
    '''Raised if 'update' call updates no doc'''

    status = 400

    def __init__(self, *, module_name):
        super().__init__(
            NoDocUpdatedException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'No documents were updated for module \'{module_name}\''


# END: update exceptions

# START: retrieve_file exceptions
class FileNotFoundException(Exception):
    '''Raised by 'retrieve_file' if failed to locate attr, or attr does not match requested file name'''

    status = 400

    def __init__(self, *, doc_id, attr_name, file_name):
        super().__init__(
            FileNotFoundException.format_msg(
                doc_id=doc_id, attr_name=attr_name, file_name=file_name
            ),
            {'doc_id': doc_id, 'attr_name': attr_name, 'file_name': file_name},
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, file_name):
        '''Formats exception message'''

        return f'Failed to find file for doc \'_id\' \'{doc_id}\', attr \'{attr_name}\', file \'{file_name}\''


class FileNotImageException(Exception):
    '''Raised by 'retrieve_file' if attemtped to generate thumbnail of non-image file'''

    status = 400

    def __init__(self, *, doc_id, attr_name, file_name, file_type):
        super().__init__(
            FileNotImageException.format_msg(
                doc_id=doc_id,
                attr_name=attr_name,
                file_name=file_name,
                file_type=file_type,
            ),
            {
                'doc_id': doc_id,
                'attr_name': attr_name,
                'file_name': file_name,
                'file_type': file_type,
            },
        )

    @staticmethod
    def format_msg(*, doc_id, attr_name, file_name, file_type):
        '''Formats exception message'''

        return f'Can\'t generate thumbnail for file for doc \'_id\' \'{doc_id}\', attr \'{attr_name}\', file \'{file_name}\', of type \'{file_type}\''


# END: retrieve_file exceptions



'''Provides collection of validation tools used by Nawah and
can be used by Nawah Apps to use with complex business-logic'''

from ._attr import validate_attr, validate_doc
from ._type import validate_type

__all__ = ['validate_attr', 'validate_doc', 'validate_type']



from typing import TYPE_CHECKING, Literal

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahEnv


async def drop(env: 'NawahEnv', collection_name: str) -> Literal[True]:
    collection = env['conn'][Config.data_name][collection_name]
    await collection.drop()
    return True



'''Provides types used in Nawah'''

from ._types import (AnalyticsEvents, AppPackage, IPQuota, NawahDoc, NawahEnv,
                     NawahEvents, NawahQueryOperAnd, NawahQueryOperOr,
                     NawahQuerySpecial, NawahQuerySpecialGroup, NawahQueryStep,
                     NawahSession, Results, ResultsArgs)

__all__ = [
    'AnalyticsEvents',
    'AppPackage',
    'IPQuota',
    'NawahDoc',
    'NawahEnv',
    'NawahEvents',
    'NawahQueryOperAnd',
    'NawahQueryOperOr',
    'NawahQuerySpecial',
    'NawahQuerySpecialGroup',
    'NawahQueryStep',
    'NawahSession',
    'Results',
    'ResultsArgs',
]



import asyncio
import datetime
import logging
import sys
import time
import traceback

import aiohttp.web

from nawah.config import Config
from nawah.enums import Event

from .._call import call
from ._common_handlers import (_not_allowed_handler, _not_found_handler,
                               _root_handler)
from ._http_handler import _http_handler
from ._utils import _close_session
from ._websocket_handler import _websocket_handler

logger = logging.getLogger('nawah')


async def _jobs_loop():
    try:
        # Connection Timeout Workflow
        logger.debug('Time to check for sessions!')
        logger.debug('Current sessions: %s', Config.sys.sessions)
        for session_id, session in Config.sys.sessions.items():
            if 'last_call' not in session.keys():
                continue
            if datetime.datetime.utcnow() > (
                session['last_call'] + datetime.timedelta(seconds=Config.conn_timeout)
            ):
                logger.debug(
                    'Session #\'%s\' with REMOTE_ADDR \'%s\' HTTP_USER_AGENT: \'%s\' is idle. '
                    'Closing',
                    session['id'],
                    session['REMOTE_ADDR'],
                    session['HTTP_USER_AGENT'],
                )
                await _close_session(session_id)
    except Exception:
        logger.error('An error occurred. Details: %s', traceback.format_exc())

    try:
        # Calls Quota Workflow - Clean-up Sequence
        logger.debug('Time to check for IPs quotas!')
        del_ip_quota = []
        for ip in Config.sys.ip_quota.keys():
            if (
                datetime.datetime.utcnow() - Config.sys.ip_quota[ip]['last_check']
            ).seconds > 59:
                logger.debug(
                    f'IP \'{ip}\' with quota \'{Config.sys.ip_quota[ip]["counter"]}\' is idle. Cleaning-up'
                )
                del_ip_quota.append(ip)
        for ip in del_ip_quota:
            del Config.sys.ip_quota[ip]
    except Exception:
        logger.error(f'An error occurred. Details: {traceback.format_exc()}')

    try:
        # Jobs Workflow
        current_time = datetime.datetime.utcnow().isoformat()[:16]
        logger.debug('Time to check for jobs!')
        for job_name, job in Config.jobs.items():
            logger.debug('Checking: %s', job_name)
            if job._disabled:
                logger.debug('-Job is disabled. Skipping')
                continue
            # Check if job is scheduled for current_time
            if current_time >= job._next_time:
                logger.debug('-Job is due, running!')
                # Update job next_time
                job._next_time = datetime.datetime.fromtimestamp(
                    job._cron_schedule.get_next(), datetime.timezone.utc
                ).isoformat()[:16]

                def job_callback(task: asyncio.Future):
                    if task.done():
                        logger.debug('-Job \'%s\' is done', job_name)

                    if task_exception := task.exception():
                        logger.error('Job \'%s\' has failed with exception', job_name)
                        logger.error('Exception details:')
                        logger.error(task_exception)
                        if job.prevent_disable:
                            logger.warning(
                                '-Detected job prevent_disable. Skipping disabling job'
                            )
                        else:
                            logger.warning('-Disabling job')
                            job._disabled = True

                job_task = asyncio.create_task(job.job(env=Config.sys.env))
                job_task.add_done_callback(job_callback)

            else:
                logger.debug('-Not yet due')
    except Exception:
        logger.error(f'An error occurred. Details: {traceback.format_exc()}')

    # [TODO] Re-implement
    # try:
    #     logger.debug('Time to check for files timeout!')
    #     files_task = asyncio.create_task(
    #         call(
    #             'file/delete',
    #             skip_events=[Event.PERM],
    #             env=Config.sys.env,
    #             query=[
    #                 {
    #                     'create_time': {
    #                         '$lt': (
    #                             datetime.datetime.utcnow()
    #                             - datetime.timedelta(seconds=Config.file_upload_timeout)
    #                         ).isoformat()
    #                     }
    #                 }
    #             ],
    #         )
    #     )
    #     # logger.debug('Files timeout results:')
    #     # logger.debug('-status: %s', files_results['status'])
    #     # logger.debug('-msg: %s', files_results['msg'])
    #     # logger.debug('-args.docs: %s', files_results["args"]['docs'])
    # except Exception:
    #     logger.error('An error occurred. Details: %s', traceback.format_exc())


def _create_error_middleware(overrides):
    @aiohttp.web.middleware
    async def error_middleware(request, handler):
        try:
            response = await handler(request)
            override = overrides.get(response.status)
            if override:
                return await override(request)
            return response
        except aiohttp.web.HTTPException as ex:
            override = overrides.get(ex.status)
            if override:
                return await override(request)
            raise

    return error_middleware


async def _web_loop():
    # Populate get_routes, post_routes
    get_routes = []
    post_routes = []
    for module_name, module in Config.modules.items():
        for func_name, func in module.funcs.items():
            if func.get_func:
                for get_args_set in func.query_attrs or [{}]:
                    if get_args_set:
                        get_args = f'/{{{"}/{".join(list(get_args_set.keys()))}}}'
                    else:
                        get_args = ''

                    get_routes.append(f'/{module_name}/{func_name}{get_args}')
            elif func.post_func:
                for post_args_set in func.query_attrs or [{}]:
                    if post_args_set:
                        post_args = f'/{{{"}/{".join(list(post_args_set.keys()))}}}'
                    else:
                        post_args = ''

                    post_routes.append(f'/{module_name}/{func_name}{post_args}')

    logger.debug(
        'Loaded modules: %s',
        {module_name: module.attrs for module_name, module in Config.modules.items()},
    )
    logger.debug(
        'Config has attrs: %s',
        {
            k: str(v)
            for k, v in Config.__dict__.items()
            if not isinstance(v, classmethod) and not k.startswith('_')
        },
    )
    logger.debug('Generated get_routes: %s', get_routes)
    logger.debug('Generated post_routes: %s', post_routes)

    app = aiohttp.web.Application()
    app.middlewares.append(
        _create_error_middleware(
            {
                404: _not_found_handler,
                405: _not_allowed_handler,
            }
        )
    )
    app.router.add_route('GET', '/', _root_handler)
    app.router.add_route('*', '/ws', _websocket_handler)
    for route in get_routes:
        app.router.add_route('GET', route, _http_handler)
    for route in post_routes:
        app.router.add_route('POST', route, _http_handler)
        app.router.add_route('OPTIONS', route, _http_handler)
    logger.info('Welcome to Nawah')
    runner = aiohttp.web.AppRunner(app)
    await runner.setup()
    site = aiohttp.web.TCPSite(runner, '0.0.0.0', Config.port)
    await site.start()
    logger.info('Serving on 0.0.0.0:%s', Config.port)
    while True:
        await asyncio.sleep(60)
        asyncio.create_task(_jobs_loop())


def _run_app():
    try:
        asyncio.run(_web_loop())
    except KeyboardInterrupt:
        if time.localtime().tm_hour >= 21 or time.localtime().tm_hour <= 4:
            msg = 'night'
        elif time.localtime().tm_hour >= 18:
            msg = 'evening'
        elif time.localtime().tm_hour >= 12:
            msg = 'afternoon'
        elif time.localtime().tm_hour >= 5:
            msg = 'morning'

        logger.info('Have a great %s!', msg)

        sys.exit()



'''Provides 'user' Module Functions callables'''

from typing import TYPE_CHECKING, Optional

from bson import ObjectId

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException
from nawah.utils import call

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


async def _read(
    skip_events: 'NawahEvents',
    env: 'NawahEnv',
    query: 'Query',
    skip_sanitise_results: Optional[bool],
) -> 'Results':
    skip_events.append(Event.PERM)
    read_results = await call(
        'base/read', module_name='user', skip_events=skip_events, env=env, query=query
    )

    if read_results['status'] != 200:
        return read_results

    # Return results as is if both sanitise_results, user_settings are to be skipped
    if skip_sanitise_results is True:
        return read_results

    # Otherwise, iterate over results and apply not skipped action
    for i, _ in enumerate(read_results['args']['docs']):
        user = read_results['args']['docs'][i]
        # Apply sanitise_results, removing hashes, if not skipped
        if skip_sanitise_results is not True:
            for user_attr_sanitise in Config.user_attrs_sanitise:
                del user[user_attr_sanitise]

    return read_results


async def _create(
    skip_events: 'NawahEvents', env: 'NawahEnv', doc: 'NawahDoc'
) -> 'Results':
    if Event.ATTRS_DOC not in skip_events:
        doc['groups'] = [ObjectId('f00000000000000000000013')]

    create_results = await call(
        'base/create', skip_events=[Event.PERM], module_name='user', env=env, doc=doc
    )

    return create_results


async def _read_privileges(env: 'NawahEnv', query: 'Query') -> 'Results':
    # Confirm _id is valid
    results = await call(
        'user/read',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['_id'][0]}],
        args={'skip_sanitise_results': True},
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
        )
    user = results['args']['docs'][0]
    for group in user['groups']:
        group_results = await call(
            'group/read', skip_events=[Event.PERM], env=env, query=[{'_id': group}]
        )
        group = group_results['args']['docs'][0]
        for privilege in group['privileges'].keys():
            if privilege not in user['privileges'].keys():
                user['privileges'][privilege] = []
            for i in range(len(group['privileges'][privilege])):
                if (
                    group['privileges'][privilege][i]
                    not in user['privileges'][privilege]
                ):
                    user['privileges'][privilege].append(
                        group['privileges'][privilege][i]
                    )
    return results


async def _add_group(
    skip_events: 'NawahEvents', env: 'NawahEnv', query: 'Query', doc: 'NawahDoc'
) -> 'Results':
    # Check for list group attr
    if isinstance(doc['group'], list):
        for i in range(0, len(doc['group']) - 1):
            await call(
                'user/add_group',
                skip_events=skip_events,
                env=env,
                query=query,
                doc={'group': doc['group'][i]},
            )
        doc['group'] = doc['group'][-1]
    # Confirm all basic args are provided
    doc['group'] = ObjectId(doc['group'])
    # Confirm group is valid
    results = await call(
        'group/read', skip_events=[Event.PERM], env=env, query=[{'_id': doc['group']}]
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
        )
    # Get user details
    results = await call('user/read', skip_events=[Event.PERM], env=env, query=query)
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
        )
    user = results['args']['docs'][0]
    # Confirm group was not added before
    if doc['group'] in user['groups']:
        raise FuncException(
            status=400,
            msg='User is already a member of the group.',
            args={'code': 'GROUP_ADDED'},
        )
    user['groups'].append(doc['group'])
    # Update the user
    results = await call(
        'user/update',
        skip_events=[Event.PERM],
        env=env,
        query=query,
        doc={'groups': user['groups']},
    )
    # if update fails, return update results
    if results['status'] != 200:
        return results
    # Check if the updated User doc belongs to current session and update it
    if env['session']['user']['_id'] == user['_id']:
        user_results = await call(
            'user/read_privileges',
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': user['_id']}],
        )
        env['session']['user'] = user_results['args']['docs'][0]

    return results


async def _delete_group(env: 'NawahEnv', query: 'Query') -> 'Results':
    # Confirm group is valid
    results = await call(
        'group/read',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['group'][0]}],
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
        )
    # Get user details
    results = await call(
        'user/read', skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
    )
    if not results['args']['count']:
        raise FuncException(
            status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
        )
    user = results['args']['docs'][0]
    # Confirm group was not added before
    if query['group'][0] not in user['groups']:
        raise FuncException(
            status=400,
            msg='User is not a member of the group.',
            args={'code': 'GROUP_NOT_ADDED'},
        )
    # Update the user
    results = await call(
        'user/update',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['_id'][0]}],
        doc={'groups': {'$del_val': [query['group'][0]]}},
    )
    # if update fails, return update results
    if results['status'] != 200:
        return results
    # Check if the updated User doc belongs to current session and update it
    if env['session']['user']['_id'] == user['_id']:
        user_results = await call(
            'user/read_privileges',
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': user['_id']}],
        )
        env['session']['user'] = user_results['args']['docs'][0]

    return results



'''Provides 'create' Data Function'''

from typing import TYPE_CHECKING

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, ResultsArgs


async def create(
    *,
    env: 'NawahEnv',
    collection_name: str,
    doc: 'NawahDoc',
) -> 'ResultsArgs':
    '''Creates 'doc' in 'collection' using connection 'conn' in 'env' '''

    collection = env['conn'][Config.data_name][collection_name]
    results = await collection.insert_one(doc)
    _id = results.inserted_id
    return {'count': 1, 'docs': [{'_id': _id}]}



'''Provides types used in Nawah'''

from typing import (TYPE_CHECKING, Any, Literal, MutableMapping,
                    MutableSequence, TypedDict, Union)

from nawah.enums import Event

if TYPE_CHECKING:
    from datetime import datetime

    from aiohttp.web import WebSocketResponse
    from motor.motor_asyncio import AsyncIOMotorClient

    from nawah.classes import Package


class Results(TypedDict):
    '''Provides type-hint for return of Nawah Function'''

    status: int
    msg: str
    args: 'ResultsArgs'


ResultsArgs = MutableMapping[str, Any]


class AppPackage(TypedDict):
    '''Provides type-hint for item of 'Config.packages' '''

    package: 'Package'
    modules: MutableSequence[str]


class AnalyticsEvents(TypedDict):
    '''Provides type-hint for 'analytics_events' App Config Attr'''

    app_conn_verified: bool
    session_conn_auth: bool
    session_user_auth: bool
    session_conn_reauth: bool
    session_user_reauth: bool
    session_conn_deauth: bool
    session_user_deauth: bool


class IPQuota(TypedDict):
    '''Provides type-hint for dict used to track IP user quota'''

    counter: int
    last_check: 'datetime'


NawahEvents = MutableSequence[Event]


class NawahEnv(TypedDict, total=False):
    '''Provides type-hint for 'Env' dict'''

    id: str
    init: bool
    conn: 'AsyncIOMotorClient'
    REMOTE_ADDR: str
    HTTP_USER_AGENT: str
    HTTP_ORIGIN: str
    client_app: str
    session: 'NawahSession'
    prev_session: MutableMapping
    prev_session_timeout: 'datetime'
    last_call: 'datetime'
    ws: 'WebSocketResponse'
    quota: 'NawahEnvQuota'
    args: MutableMapping[str, Any]


class NawahEnvQuota(TypedDict):
    '''Provides type-hint for 'quota' of 'Env' dict'''

    counter: int
    last_check: 'datetime'


class NawahSession(TypedDict):
    '''Provieds type-hint for 'session' of 'Env' dict'''

    _id: str
    user: MutableMapping
    groups: MutableSequence[str]
    host_add: str
    user_agent: str
    expiry: str
    token_hash: str
    create_time: str


class NawahQuerySpecialGroup(TypedDict):
    '''Provides type-hint for '$group' in 'NawahQuery' '''

    by: str
    count: int


class NawahQuerySpecialGeoNear(TypedDict):
    '''Provides type-hint for '$geo_near' in 'NawahQuery' '''

    val: str
    attr: str
    dist: int


# Following TypedDict type can't be defined as class as keys include $
NawahQuerySpecial = TypedDict(
    'NawahQuerySpecial',
    {
        '$search': str,
        '$sort': MutableMapping[str, Literal[1, -1]],
        '$skip': int,
        '$limit': int,
        '$extn': Union[bool, MutableSequence[str]],
        '$attrs': MutableSequence[str],
        '$group': MutableSequence[NawahQuerySpecialGroup],
        '$geo_near': NawahQuerySpecialGeoNear,
    },
    total=False,
)

NawahQueryOperEq = TypedDict('NawahQueryOperEq', {'$eq': Any})
NawahQueryOperNe = TypedDict('NawahQueryOperNe', {'$ne': Any})
NawahQueryOperGt = TypedDict('NawahQueryOperGt', {'$gt': Union[int, float, str]})
NawahQueryOperGte = TypedDict('NawahQueryOperGte', {'$gte': Union[int, float, str]})
NawahQueryOperLt = TypedDict('NawahQueryOperLt', {'$lt': Union[int, float, str]})
NawahQueryOperLte = TypedDict('NawahQueryOperLte', {'$lte': Union[int, float, str]})
NawahQueryOperAll = TypedDict('NawahQueryOperAll', {'$all': list[Any]})
NawahQueryOperIn = TypedDict('NawahQueryOperIn', {'$in': list[Any]})
NawahQueryOperNin = TypedDict('NawahQueryOperNin', {'$nin': list[Any]})
NawahQueryOperRegex = TypedDict('NawahQueryOperRegex', {'$regex': str})
NawahQueryStep = dict[
    str,
    Union[
        NawahQueryOperEq,
        NawahQueryOperNe,
        NawahQueryOperGt,
        NawahQueryOperGte,
        NawahQueryOperLt,
        NawahQueryOperLte,
        NawahQueryOperAll,
        NawahQueryOperIn,
        NawahQueryOperNin,
        NawahQueryOperRegex,
    ],
]

NawahQueryOperOr = TypedDict('NawahQueryOperOr', {'$or': list[NawahQueryStep]})
NawahQueryOperAnd = TypedDict('NawahQueryOperAnd', {'$and': list[NawahQueryStep]})

NawahQueryIndex = TypedDict(
    'NawahQueryIndex', {'$index': dict[str, list[tuple[str, Any]]]}
)

NawahDoc = MutableMapping[str, Any]



'''Provides 'validate_type' Utility'''

import inspect
import logging
import re
from typing import TYPE_CHECKING, Any

from nawah.classes import Attr
from nawah.config import Config
from nawah.enums import AttrType
from nawah.exceptions import (InvalidAttrTypeArgException,
                              InvalidAttrTypeException,
                              InvalidAttrTypeRefException)

from .._shared import ATTRS_TYPES_ARGS

if TYPE_CHECKING:

    from .._shared import AttrTypeArgValidation

logger = logging.getLogger('nawah')


ATTR_TYPE_ARG_TYPE_MAP = {
    'Attr': lambda arg_val: isinstance(arg_val, Attr),
    'bool': lambda arg_val: isinstance(arg_val, bool),
    'callable': lambda arg_val: callable(arg_val)
    or inspect.iscoroutinefunction(arg_val),
    'dict': lambda arg_val: isinstance(arg_val, dict),
    'float': lambda arg_val: isinstance(arg_val, float),
    'int': lambda arg_val: isinstance(arg_val, int),
    'list': lambda arg_val: isinstance(arg_val, list),
    'str': lambda arg_val: isinstance(arg_val, str),
    'union': lambda arg_val: isinstance(arg_val, list),
}


def validate_type(attr_type: 'Attr', skip_type: bool = False):
    '''Validates Attr Type objects. If failed, raises 'InvalidAttrTypeException' '''
    # Skip validating Attr Type if it is already validated
    if skip_type and attr_type.valid:
        return

    if attr_type.type not in AttrType:
        raise InvalidAttrTypeException(attr_type=attr_type)

    if attr_type.type == AttrType.TYPE:
        if skip_type:
            Config.sys.type_attrs.append(attr_type)
            return

        if isinstance(attr_type.args['type'], str):
            if attr_type.args['type'] not in Config.types:
                raise InvalidAttrTypeRefException(attr_type=attr_type)
            if not callable(Config.types[attr_type.args['type']]):
                raise InvalidAttrTypeRefException(attr_type=attr_type)
            # Assign new Attr Type Arg for shorthand calling the TYPE function
            attr_type.args['func'] = Config.types[attr_type.args['type']]
        else:
            if not inspect.iscoroutinefunction(attr_type.args['type']):
                raise InvalidAttrTypeRefException(attr_type=attr_type)
            # Assign new Attr Type Arg for shorthand calling the TYPE function
            attr_type.args['func'] = attr_type.args['type']

        attr_type.valid = True
        return

    for arg in (attr_type_args := ATTRS_TYPES_ARGS[attr_type.type.name]):
        try:
            if 'required' in attr_type_args[arg] or arg in attr_type.args:
                _validate_type_arg(
                    arg_name=arg,
                    arg_type=attr_type_args[arg],
                    arg_val=attr_type.args[arg],
                    skip_type=skip_type,
                )
        except InvalidAttrTypeArgException as e:
            raise InvalidAttrTypeException(attr_type=attr_type) from e

    attr_type.valid = True


def _validate_type_arg(
    arg_name: str, arg_type: 'AttrTypeArgValidation', arg_val: Any, skip_type: bool
):
    '''Validates Attr Type objects args. If failed, raises 'InvalidAttrTypeArgException' '''

    if not ATTR_TYPE_ARG_TYPE_MAP[arg_type['type']](arg_val):
        if 'required' in arg_type:
            raise InvalidAttrTypeArgException(
                arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
            )

    if arg_val is None:
        return

    if arg_type['type'] == 'dict':
        for key, val in arg_val.items():
            _validate_type_arg(
                arg_name=f'{arg_name}.{key}',
                arg_type=arg_type['dict_key_type'],
                arg_val=key,
                skip_type=skip_type,
            )
            _validate_type_arg(
                arg_name=f'{arg_name}.{key}.val',
                arg_type=arg_type['dict_val_type'],
                arg_val=val,
                skip_type=skip_type,
            )
        return

    if arg_type['type'] == 'list':
        if arg_type['list_len_range'] is not None:
            if not (
                arg_type['list_len_range'][0]
                <= len(arg_val)
                < arg_type['list_len_range'][1]
            ):
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
        for i, _ in enumerate(arg_val):
            _validate_type_arg(
                arg_name=f'{arg_name}.{i}',
                arg_type=arg_type['list_item_type'],
                arg_val=arg_val[i],
                skip_type=skip_type,
            )
        return

    if arg_type['type'] == 'str':
        if 'str_pattern' in arg_type:
            if not re.match(f'^{arg_type["str_pattern"]}$', arg_val):
                raise InvalidAttrTypeArgException(
                    arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
                )
        return

    if arg_type['type'] == 'float':
        if isinstance(arg_val, (int, float)):
            return

    if arg_type['type'] == 'union':
        for union_type in arg_type['union_types']:
            # For union arg type, loop over all allowed types, validate
            # arg_val against all of them, and break when one successes,
            # Otherwise raise InvalidAttrTypeArgException
            try:
                _validate_type_arg(
                    arg_name=arg_name,
                    arg_type=union_type,
                    arg_val=arg_val,
                    skip_type=skip_type,
                )
                return
            except InvalidAttrTypeArgException:
                pass

    if arg_type['type'] == 'Attr':
        validate_type(attr_type=arg_val, skip_type=skip_type)
        return

    raise InvalidAttrTypeArgException(
        arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
    )



'''Provides 'call' Utility'''

import asyncio
import copy
import datetime
import inspect
import json
import logging
import re
import sys
import traceback
from typing import (TYPE_CHECKING, Any, Awaitable, Callable, MutableMapping,
                    MutableSequence, Union, cast)

from bson import ObjectId

from nawah.classes import Query, Var, app_encoder
from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import (FuncException, InvalidAttrException,
                              InvalidCallEndpointException,
                              InvalidDocAttrException, InvalidDocException,
                              InvalidFuncException, InvalidModuleException,
                              InvalidQueryAttrException, InvalidQueryException,
                              MissingDocAttrException,
                              MissingQueryAttrException)

from ._cache import _get_cache, _set_cache, reset_cache_channel
from ._check_permissions import check_permissions as check_permissions_utility
from ._val import camel_to_upper, var_value
from ._validate import validate_attr

if TYPE_CHECKING:
    from nawah.classes import Attr
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


logger = logging.getLogger('nawah')


async def call(
    endpoint: str,
    /,
    *,
    module_name: str = None,
    skip_events: 'NawahEvents' = None,
    env: 'NawahEnv' = None,
    query: 'Query' = None,
    doc: 'NawahDoc' = None,
    args: MutableMapping[str, Any] = None,
) -> 'Results':
    '''Checks validity of an endpoint and calls Nawah Function callable at endpoint,
    returning the couroutine of the callable. If endpoint points to non-existent
    Nawah Function, raises 'InvalidFuncException' '''

    if not re.match(r'^[a-z_]+\/[a-z_]+$', endpoint):
        raise InvalidCallEndpointException(endpoint=endpoint)

    endpoint_module, endpoint_func = endpoint.split('/')

    try:
        module = Config.modules[endpoint_module]
    except KeyError as e:
        raise InvalidModuleException(module_name=endpoint_module) from e

    try:
        func = module.funcs[endpoint_func]
    except KeyError as e:
        raise InvalidFuncException(
            module_name=endpoint_module, func_name=endpoint_func
        ) from e

    # Set defaults for kwargs
    module_name = module_name or endpoint_module
    skip_events = skip_events or []
    query = query or Query()
    doc = doc or {}
    env = env or {}
    args = args or {}

    # Convert query to Query object
    if not isinstance(query, Query):
        logger.warning('Call is using deprecated query format: %s', query)
        query_special = {}
        for step in query:
            if not isinstance(step, dict):
                continue
            for attr_name, attr_val in step.items():
                if isinstance(attr_name, str) and attr_name[0] == '$':
                    query_special[attr_name] = attr_val
        query = Query(pipe=_convert_deprecated_query(query), special=query_special)

    # Check conditions for call checks
    check_permissions = Event.PERM not in skip_events
    check_attrs_query = Event.ATTRS_QUERY not in skip_events
    check_attrs_doc = Event.ATTRS_DOC not in skip_events
    check_cache = Event.CACHE not in skip_events
    check_analytics = Event.ANALYTICS not in skip_events

    if check_permissions:
        query_mod, doc_mod = check_permissions_utility(func=func, env=env)
        # Use return of check_permissions_utility to update query, doc
        query.pipe.append(
            _process_query_mod(
                query=query,
                query_mod=query_mod,
                env=env,
                doc=doc,
            )
        )
        doc.update(
            _process_doc_mod(
                doc=doc,
                doc_mod=doc_mod,
                env=env,
                query=query,
            )
        )

    if check_attrs_query:
        func.query_attrs = cast(MutableSequence, func.query_attrs)
        _check_query_attrs(query=query, query_attrs=func.query_attrs)

    if check_attrs_doc:
        func.doc_attrs = cast(MutableSequence, func.doc_attrs)
        _check_doc_attrs(doc=doc, doc_attrs=func.doc_attrs)

    if check_cache:
        cache_key, call_cache = await _get_cache(
            func=func,
            skip_events=skip_events,
            env=env,
            query=query,
        )

        if call_cache:
            return call_cache

    if check_analytics:
        # [TODO] Implement
        pass

    try:
        func_callable = cast(Callable[..., Awaitable['Results']], func.callable)
        kwargs: MutableMapping = {
            'func': func,
            'module_name': module_name,
            'skip_events': skip_events,
            'env': env,
            'query': query,
            'doc': doc,
        }

        # Iterate over call_args to set values
        for arg_name, arg_type in (func.call_args or {}).items():
            if arg_name not in args:
                kwargs[arg_name] = None
                continue

            try:
                validate_attr(
                    mode='create',
                    attr_name=arg_name,
                    attr_type=arg_type,
                    attr_val=args[arg_name],
                )
            except InvalidAttrException as e:
                raise FuncException(
                    status=500,
                    msg=f'Invalid value for \'Call Arg\' \'{arg_name}\' of \'{module_name}\''
                    f'.\'{func.name}\'',
                    args={'code': 'INVALID_CALL_ARG'},
                ) from e
            # Call Arg value is value, append it to kwargs to be passed to Func callable
            kwargs[arg_name] = args[arg_name]
            # Delete arg_val from args to filter out invalid args
            del args[arg_name]

        # If after iterating over call_args, we still have values in args, declare them invalid
        if args:
            arg_name = list(args)[0]
            raise FuncException(
                status=500,
                msg=f'Function of \'{module_name}\'.\'{func.name}\' does not define \'{arg_name}\' '
                'in \'call_args\'',
                args={'code': 'UNKNOWN_CALL_ARG'},
            )

        results = await func_callable(
            **{
                param: kwargs[param]
                for param in inspect.signature(func_callable).parameters
                if param in kwargs
            }
        )
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()

        logger.error(
            'Callable for \'%s\' failed with Exception of type: %s', endpoint, type(e)
        )
        # [TODO] Check its validity
        logger.debug('Error traceback:')
        for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
            logger.debug('- %s', line)

        if isinstance(e, FuncException):
            raise e

        raise FuncException(
            status=getattr(e, 'status', 500),
            msg=e.args[0] if e.args else 'Unexpected error has occurred',
            args={'code': camel_to_upper(e.__class__.__name__)},
        ) from e

    if 'args' in results and 'session' in results['args']:
        env['session'] = json.loads(app_encoder.encode(results['args']['session']))

    if check_cache and cache_key:
        results['args']['cache_key'] = cache_key
        if 'cache_time' not in results['args']:
            logger.debug(
                'Results generated with \'cache_key\'. Calling \'_set_cache\'.'
            )
            results['args']['cache_time'] = datetime.datetime.utcnow().isoformat()
        # [TODO] Add callback to handle errors
        asyncio.create_task(_set_cache(func=func, cache_key=cache_key, results=results))

    if func.cache_channels_reset:
        for channel in func.cache_channels_reset:
            # [TODO] Add callback to handle errors
            asyncio.create_task(reset_cache_channel(channel))

    return results


def _process_query_mod(
    *,
    query_mod: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ],
    env: 'NawahEnv',
    query: 'Query',
    doc: 'NawahDoc',
):
    if not query_mod:
        return {}

    query_mod_processed: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ]

    if isinstance(query_mod, list):
        query_mod_processed = []
        for query_mod_child in query_mod:
            query_mod_processed.append(
                _process_query_mod(
                    query_mod=query_mod_child,
                    env=env,
                    query=query,
                    doc=doc,
                )
            )

    elif isinstance(query_mod, dict):
        query_mod_processed = {}
        for attr_name, attr_val in query_mod.items():
            if isinstance(attr_val, (list, dict)):
                query_mod_processed[attr_name] = _process_query_mod(
                    query_mod=attr_val,
                    env=env,
                    query=query,
                    doc=doc,
                )
            elif isinstance(attr_val, Var):
                query_mod_processed[attr_name] = var_value(
                    attr_val, env=env, doc=doc, locale=env['session']['user']['locale']
                )
            else:
                query_mod_processed[attr_name] = copy.deepcopy(attr_val)
    else:
        raise Exception('Unexpected non-list, non-dict value for \'query_mod\'')

    return query_mod_processed


def _process_doc_mod(
    *,
    doc_mod: Union[MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]],
    env: 'NawahEnv',
    query: 'Query',
    doc: 'NawahDoc',
):
    if not doc_mod:
        return {}

    doc_mod_processed: Union[
        MutableMapping[str, Any], MutableSequence[MutableMapping[str, Any]]
    ]

    if isinstance(doc_mod, list):
        doc_mod_processed = []
        for doc_mod_child in doc_mod:
            doc_mod_processed.append(
                _process_doc_mod(
                    doc_mod=doc_mod_child,
                    env=env,
                    query=query,
                    doc=doc,
                )
            )

    elif isinstance(doc_mod, dict):
        doc_mod_processed = {}
        for attr_name, attr_val in doc_mod.items():
            if isinstance(attr_val, (list, dict)):
                doc_mod_processed[attr_name] = _process_doc_mod(
                    doc_mod=attr_val,
                    env=env,
                    query=query,
                    doc=doc,
                )
            elif isinstance(attr_val, Var):
                doc_mod_processed[attr_name] = var_value(
                    attr_val, env=env, doc=doc, locale=env['session']['user']['locale']
                )
            else:
                doc_mod_processed[attr_name] = copy.deepcopy(attr_val)
    else:
        raise Exception('Unexpected non-list, non-dict value for \'doc_mod\'')

    return doc_mod_processed


def _check_query_attrs(
    *,
    query: 'Query',
    query_attrs: MutableSequence[MutableMapping[str, 'Attr']],
):
    if not query_attrs:
        return

    query_attrs_sets: MutableSequence[MutableMapping[str, str]] = []

    for query_attrs_set in query_attrs:
        query_attrs_sets.append({})
        try:
            for attr_name, attr_type in query_attrs_set.items():
                logger.warning(
                    'Detected attr for Func.query_attrs with no oper: %s', attr_name
                )
                # [TODO] Remove when nawah query is obsoleted
                if ':' not in attr_name:
                    attr_name = f'{attr_name}:$eq'
                if attr_name not in query:
                    query_attrs_sets[-1][attr_name] = 'missing'
                    raise MissingQueryAttrException(attr_name=attr_name)

                for i, attr_val in enumerate(query[attr_name]):
                    try:
                        query[attr_name][i] = validate_attr(
                            mode='create',
                            attr_name=attr_name,
                            attr_type=attr_type,
                            attr_val=attr_val,
                        )
                    except InvalidAttrException as e:
                        query_attrs_sets[-1][attr_name] = 'invalid'
                        raise InvalidQueryAttrException(
                            attr_name=attr_name,
                            attr_type=attr_type,
                            val_type=type(attr_val),
                        ) from e

                query_attrs_sets[-1][attr_name] = 'valid'
            # If looped successfully over complete set, return to indicate valid Query
            return
        except (InvalidAttrException, MissingQueryAttrException):
            # If exception occur, pass it to allow checking all sets
            pass

    # If all sets are checked but failed to return, rase InvalidQueryException
    raise InvalidQueryException(query_attrs_sets=query_attrs_sets)


def _check_doc_attrs(
    *,
    doc: 'NawahDoc',
    doc_attrs: MutableSequence[MutableMapping[str, 'Attr']],
):
    if not doc_attrs:
        return

    doc_attrs_sets: MutableSequence[MutableMapping[str, str]] = []

    for doc_attrs_set in doc_attrs:
        doc_attrs_sets.append({})
        try:
            for attr_name, attr_type in doc_attrs_set.items():
                if attr_name not in doc:
                    doc_attrs_sets[-1][attr_name] = 'missing'
                    raise MissingDocAttrException(attr_name=attr_name)

                try:
                    doc[attr_name] = validate_attr(
                        mode='create',
                        attr_name=attr_name,
                        attr_type=attr_type,
                        attr_val=doc[attr_name],
                    )
                except InvalidAttrException as e:
                    doc_attrs_sets[-1][attr_name] = 'invalid'
                    raise InvalidDocAttrException(
                        attr_name=attr_name,
                        attr_type=attr_type,
                        val_type=type(doc[attr_name]),
                    ) from e

                doc_attrs_sets[-1][attr_name] = 'valid'
            # If looped successfully over complete set, return to indicate valid Doc
            return
        except (InvalidAttrException, MissingDocAttrException):
            # If exception occur, pass it to allow checking all sets
            pass

    # If all sets are checked but failed to return, rase InvalidDocException
    raise InvalidDocException(doc_attrs_sets=doc_attrs_sets)


def _convert_deprecated_query(query):
    shadow_query = []
    for step in query:
        if isinstance(step, list):
            shadow_query.append({'$or': _convert_deprecated_query(step)})

        elif isinstance(step, dict):
            for attr, val in step.items():
                if attr[0] == '$':
                    continue
                if isinstance(val, dict) and (
                    len(val) and list(val)[0].startswith('$')
                ):
                    shadow_step = {attr: val}
                else:
                    shadow_step = {attr: {'$eq': val}}

                if attr == '_id':
                    if isinstance(
                        shadow_step['_id'][list(shadow_step['_id'])[0]], list
                    ):
                        shadow_step['_id'][list(shadow_step['_id'])[0]] = [
                            ObjectId(item)
                            for item in shadow_step['_id'][list(shadow_step['_id'])[0]]
                        ]
                    else:
                        shadow_step['_id'][list(shadow_step['_id'])[0]] = ObjectId(
                            shadow_step['_id'][list(shadow_step['_id'])[0]]
                        )

                shadow_query.append(shadow_step)

    return shadow_query



'''Provides exceptions related to Query'''

from typing import Any, Literal


class InvalidQueryStepTypeException(Exception):
    '''Raised by \'validate_query_step\' if type of Query step is not 'dict' '''

    status = 400

    def __init__(self, *, step_type: Any):
        super().__init__(
            InvalidQueryStepTypeException.format_msg(step_type=step_type),
            {'step_type': step_type},
        )

    @staticmethod
    def format_msg(*, step_type):
        '''Formats exception message'''

        return (
            f'Query step type should be of type \'dict\'. Got \'{step_type}\' instead'
        )


class InvalidQueryStepLenException(Exception):
    '''Raised by \'validate_query_step\' if Query step has more than one item'''

    status = 400

    def __init__(self, *, step_items: tuple[str]):
        super().__init__(
            InvalidQueryStepLenException.format_msg(step_items=step_items),
            {'step_items': step_items},
        )

    @staticmethod
    def format_msg(*, step_items):
        '''Formats exception message'''

        return f'Query step can only have one item. Got \'{step_items}\' instead'


class InvalidQueryStepAttrTypeException(Exception):
    '''Raised by \'validate_query_step\' if type of Query step attr is not 'dict' '''

    status = 400

    def __init__(self, *, step_attr_type: Any):
        super().__init__(
            InvalidQueryStepAttrTypeException.format_msg(step_attr_type=step_attr_type),
            {'step_attr_type': step_attr_type},
        )

    @staticmethod
    def format_msg(*, step_attr_type):
        '''Formats exception message'''

        return f'Query step attr value should be of type \'dict\'. Got \'{step_attr_type}\' instead'


class InvalidQueryStepAttrLenException(Exception):
    '''Raised by \'validate_query_step\' if Query step attr has more than one item'''

    status = 400

    def __init__(self, *, step_attr_items: tuple[str]):
        super().__init__(
            InvalidQueryStepAttrLenException.format_msg(
                step_attr_items=step_attr_items
            ),
            {'step_attr_items': step_attr_items},
        )

    @staticmethod
    def format_msg(*, step_attr_items):
        '''Formats exception message'''

        return (
            f'Query step attr can only have one item. Got \'{step_attr_items}\' instead'
        )


class InvalidQueryOperTypeException(Exception):
    '''Raised by \'validate_query_step\' if value of Query step attr is invalid'''

    status = 400

    def __init__(
        self,
        *,
        attr_name: str,
        attr_oper: Literal[
            '$ne',
            '$eq',
            '$gt',
            '$gte',
            '$lt',
            '$lte',
            '$all',
            '$in',
            '$nin',
            '$regex',
        ],
        attr_type: Any,
        attr_val: Any,
    ):
        super().__init__(
            InvalidQueryOperTypeException.format_msg(
                attr_name=attr_name,
                attr_oper=attr_oper,
                attr_type=attr_type,
                attr_val=attr_val,
            ),
            {
                'attr_name': attr_name,
                'attr_oper': attr_oper,
                'attr_type': attr_type,
                'attr_val': attr_val,
            },
        )

    @staticmethod
    def format_msg(*, attr_name, attr_oper, attr_type, attr_val):
        '''Formats exception message'''

        return f'Invalid value for Query Arg \'{attr_name}\' with Query Arg Oper \'{attr_oper}\' expecting type \'{attr_type}\' but got \'{attr_val}\''


class UnknownQueryOperException(Exception):
    '''Raised by \'validate_query_step\' if an unknown \'Query Arg\' is detected'''

    status = 400

    def __init__(
        self,
        *,
        attr_name: str,
        attr_oper: Literal[
            '$ne',
            '$eq',
            '$gt',
            '$gte',
            '$lt',
            '$lte',
            '$all',
            '$in',
            '$nin',
            '$regex',
        ],
    ):
        super().__init__(
            UnknownQueryOperException.format_msg(
                attr_name=attr_name, attr_oper=attr_oper
            ),
            {'attr_name': attr_name, 'attr_oper': attr_oper},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_oper):
        '''Formats exception message'''

        return f'Unknown Query Arg Oper \'{attr_oper}\' for Query Arg \'{attr_name}\''



'''Provides 'Analytics' dataclass'''

from dataclasses import dataclass
from typing import Callable


@dataclass(kw_only=True)
class Analytics:
    '''Analytics dataclass serves role of defining an Analytics Set, which
    instructs Nawah to record analytics event for calls which 'condition'
    does not raise Exception for, formatted as 'props' '''

    condition: Callable
    props: Callable

    def __post_init__(self):
        pass



import copy
import logging
from typing import (TYPE_CHECKING, Any, MutableMapping, Optional, Tuple, Union,
                    cast)

from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.classes import Attr, Query
    from nawah.types import (NawahQueryOperAnd, NawahQueryOperOr,
                             NawahQuerySpecialGroup, NawahQueryStep)

logger = logging.getLogger('nawah')


def _compile_query(
    *,
    attrs: MutableMapping[str, 'Attr'],
    query: 'Query',
) -> Tuple[
    Optional[int],
    Optional[int],
    MutableMapping[str, int],
    Optional[list['NawahQuerySpecialGroup']],
    list[Any],
]:
    aggregate_prefix: list[Any] = []
    aggregate_suffix: list[Any] = []
    aggregate_query: list[Any] = [{'$match': {'$and': []}}]
    aggregate_match = aggregate_query[0]['$match']['$and']
    skip: Optional[int] = None
    limit: Optional[int] = None
    sort: MutableMapping[str, int] = {'_id': -1}
    group: Optional[list['NawahQuerySpecialGroup']] = None
    logger.debug('attempting to process query: %s', query)

    query = copy.deepcopy(query)

    # Update variables per Doc Mode
    if '__deleted' not in query or query['__deleted'] is False:
        aggregate_prefix.append({'$match': {'__deleted': {'$exists': False}}})
    else:
        # This condition is expanded to allow __deleted = True, __deleted = None to have
        # del query[__deleted] be applied to both conditions
        if query['__deleted'][0] is True:
            aggregate_prefix.append({'$match': {'__deleted': {'$exists': True}}})
        del query['__deleted'][0]

    if '__create_draft' not in query or query['__create_draft'] is False:
        aggregate_prefix.append({'$match': {'__create_draft': {'$exists': False}}})
    else:
        if query['__create_draft'][0] is True:
            aggregate_prefix.append({'$match': {'__create_draft': {'$exists': True}}})
        del query['__create_draft'][0]

    if ('__update_draft' not in query and '__update_draft:$ne' not in query) or query[
        '__update_draft'
    ] is False:
        query_update_draft = False
        aggregate_prefix.append({'$match': {'__update_draft': {'$exists': False}}})
    else:
        query_update_draft = True
        aggregate_prefix.append({'$match': {'__update_draft': {'$exists': True}}})
        if '__update_draft' in query and isinstance(
            query['__update_draft'][0], ObjectId
        ):
            aggregate_prefix.append(
                {'$match': {'__update_draft': query['__update_draft'][0]}}
            )
        elif '__update_draft:$ne' in query and query['__update_draft:$ne'][0] is False:
            aggregate_prefix.append({'$match': {'__update_draft': {'$ne': False}}})
        try:
            del query['__update_draft'][0]
        except:
            del query['__update_draft:$ne'][0]

    # Update variables per Query Special Args
    if '$skip' in query:
        skip = query['$skip']
    if '$limit' in query:
        limit = query['$limit']
    if '$sort' in query:
        sort = query['$sort']
    if '$group' in query:
        group = query['$group']
    if '$search' in query:
        aggregate_prefix.insert(0, {'$match': {'$text': {'$search': query['$search']}}})
        project_query: MutableMapping[str, Any] = {
            attr: '$' + attr for attr in attrs.keys()
        }
        project_query['_id'] = '$_id'
        project_query['__score'] = {'$meta': 'textScore'}
        aggregate_suffix.append({'$project': project_query})
        aggregate_suffix.append({'$match': {'__score': {'$gt': 0.5}}})
    if '$geo_near' in query:
        aggregate_prefix.insert(
            0,
            {
                '$geoNear': {
                    'near': {
                        'type': 'Point',
                        'coordinates': query['$geo_near']['val'],
                    },
                    'distanceField': query['$geo_near']['attr'] + '.__distance',
                    'maxDistance': query['$geo_near']['dist'],
                    'spherical': True,
                }
            },
        )

    aggregate_match = query.pipe or []

    for step in aggregate_match:
        _compile_query_step(
            aggregate_prefix=aggregate_prefix,
            aggregate_suffix=aggregate_suffix,
            attrs=attrs,
            step=step,
        )

    if '$attrs' in query and isinstance(query['$attrs'], list):
        group_query: dict[str, Any] = {
            '_id': '$_id',
            **{
                attr: {'$first': f'${attr}'}
                for attr in query['$attrs']
                if attr in attrs.keys()
            },
        }
        # We need to expose __update_draft value if it is queried as this refers to the original
        # doc to be updated
        if query_update_draft:
            group_query['__update_draft'] = {'$first': '$__update_draft'}
        aggregate_suffix.append({'$group': group_query})
    else:
        group_query = {
            '_id': '$_id',
            **{attr: {'$first': f'${attr}'} for attr in attrs.keys()},
        }
        # We need to expose __update_draft value if it is queried as this refers to the original
        # doc to be updated
        if query_update_draft:
            group_query['__update_draft'] = {'$first': '$__update_draft'}
        aggregate_suffix.append({'$group': group_query})

    logger.debug(
        'processed query, aggregate_prefix:%s, aggregate_suffix:%s, aggregate_match:%s',
        aggregate_prefix,
        aggregate_suffix,
        aggregate_match,
    )
    if len(aggregate_match) == 1:
        aggregate_query = [{'$match': aggregate_match[0]}]
    elif len(aggregate_match) == 0:
        aggregate_query = []

    aggregate_query = aggregate_prefix + aggregate_query + aggregate_suffix
    return (skip, limit, sort, group, aggregate_query)


def _compile_query_step(
    *,
    aggregate_prefix: list[Any],
    aggregate_suffix: list[Any],
    attrs: MutableMapping[str, 'Attr'],
    step: Union['NawahQueryStep', 'NawahQueryOperOr', 'NawahQueryOperAnd'],
) -> None:
    for attr_name, attr_val in step.items():
        if isinstance(attr_val, list):
            for child_attr_val in attr_val:
                _compile_query_step(
                    aggregate_prefix=aggregate_prefix,
                    aggregate_suffix=aggregate_suffix,
                    attrs=attrs,
                    step=child_attr_val,
                )
            return

        if '.' not in attr_name:
            continue

        root_attr_name = attr_name.split('.')[0]

        if root_attr_name not in attrs or not attrs[root_attr_name].extn:
            continue

        # [TODO] Check if this works with EXTN as Attr Type TYPE
        # Don't attempt to extn attr that is already extended
        lookup_query = False
        for stage in aggregate_prefix:
            if '$lookup' in stage.keys() and stage['$lookup']['as'] == root_attr_name:
                lookup_query = True
                break
        if not lookup_query:
            extn_collection = Config.modules[
                attrs[root_attr_name].extn.module
            ].collection
            aggregate_prefix.append(
                {'$addFields': {root_attr_name: {'$toObjectId': root_attr_name}}}
            )
            aggregate_prefix.append(
                {
                    '$lookup': {
                        'from': extn_collection,
                        'localField': root_attr_name,
                        'foreignField': '_id',
                        'as': root_attr_name,
                    }
                }
            )
            aggregate_prefix.append({'$unwind': root_attr_name})
            group_query: MutableMapping[str, Any] = {
                attr: {'$first': f'${attr}'} for attr in attrs.keys()
            }
            group_query[root_attr_name] = {'$first': f'${root_attr_name}._id'}
            group_query['_id'] = '$_id'
            aggregate_suffix.append({'$group': group_query})



'''Provides 'setting' Nawah Core Module'''

from nawah.classes import Attr, Diff, Extn, Func, Module, Perm, Var

from ._callables import _create, _update

setting = Module(
    name='setting',
    desc='\'setting\' module module provides data type and controller for settings in Nawah '
    'eco-system. This is used by \'User\' module tp provide additional user-wise settings. It '
    'also allows for global-typed settings',
    collection='setting_docs',
    attrs={
        'user': Attr.ID(desc='\'_id\' of \'User\' doc the doc belongs to'),
        'var': Attr.STR(
            desc='Name of the setting. This is unique for every \'user\' in the module'
        ),
        'val': Attr.ANY(desc='Value of the setting'),
        'val_type': Attr.ATTR(),
        'type': Attr.LITERAL(
            desc='Type of the setting. This sets whether setting is global, or belong to user, '
            'and whether use can update it or not',
            literal=['global', 'user', 'user_sys'],
        ),
    },
    diff=Diff(condition=lambda *_: True),
    unique_attrs=[('user', 'var', 'type')],
    extns={
        'val': Extn(
            module=Var.DOC('val.__extn.module'),
            attrs=Var.DOC('val.__extn.attrs'),
            force=Var.DOC('val.__extn.force'),
        ),
    },
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='admin', query_mod={'$limit': 1}),
                Perm(
                    privilege='read',
                    query_mod={
                        'user': Var.SESSION('user'),
                        # 'type': Attr.TYPE(type=attr_query_mod_type),
                        '$limit': 1,
                    },
                ),
            ],
            query_attrs=[
                {
                    '_id': Attr.ID(),
                    'type': Attr.LITERAL(literal=['global', 'user', 'user_sys']),
                },
                {
                    'var': Attr.STR(),
                    'type': Attr.LITERAL(literal=['global']),
                },
                {
                    'var': Attr.STR(),
                    'user': Attr.ID(),
                    'type': Attr.LITERAL(literal=['user', 'user_sys']),
                },
            ],
        ),
        'create': Func(
            permissions=[
                Perm(privilege='admin'),
                Perm(privilege='create', doc_mod={'type': 'user'}),
            ],
            call_args={'raise_no_success': Attr.BOOL()},
            callable=_create,
        ),
        'update': Func(
            permissions=[
                Perm(privilege='admin', query_mod={'$limit': 1}),
                Perm(
                    privilege='update',
                    query_mod={
                        'type': 'user',
                        'user': Var.SESSION('user'),
                        '$limit': 1,
                    },
                    doc_mod={'var': None, 'val_type': None, 'type': None},
                ),
            ],
            query_attrs=[
                {
                    '_id': Attr.ID(),
                    'type': Attr.LITERAL(literal=['global', 'user', 'user_sys']),
                },
                {
                    'var': Attr.STR(),
                    'type': Attr.LITERAL(literal=['global']),
                },
                {
                    'var': Attr.STR(),
                    'user': Attr.ID(),
                    'type': Attr.LITERAL(literal=['user', 'user_sys']),
                },
            ],
            call_args={'raise_no_success': Attr.BOOL()},
            callable=_update,
        ),
        'delete': Func(
            permissions=[Perm(privilege='admin', query_mod={'$limit': 1})],
            query_attrs=[{'_id': Attr.ID()}, {'var': Attr.STR()}],
        ),
        'retrieve_file': Func(
            permissions=[Perm(privilege='*', query_mod={'type': 'global'})],
            get_func=True,
        ),
    },
)



import logging
from typing import Any, Dict

from motor.motor_asyncio import AsyncIOMotorClient

from nawah.config import Config

logger = logging.getLogger('nawah')


def create_conn() -> AsyncIOMotorClient:
    connection_config: Dict[str, Any] = {'ssl': Config.data_ssl}
    conn = AsyncIOMotorClient(Config.data_server, **connection_config, connect=True)
    return conn



'''Provides 'session' Nawah Core Module'''

from nawah.classes import Attr, Default, Extn, Func, Module, Perm, Var

from ._callables import _auth, _reauth, _signout

session = Module(
    name='session',
    desc='\'session\' module provides data type and controller for sessions in Nawah eco-system. '
    'CRUD methods of the module are supposed to used for internal calls only, while methods '
    '\'auth\', \'reauth\', and \'signout\' are available for use by API as well as internal calls '
    'when needed',
    collection='session_docs',
    attrs={
        'user': Attr.ID(desc='\'_id\' of \'User\' doc the doc belongs to'),
        'groups': Attr.LIST(
            desc='List of \'_id\' for every group the session is authenticated against. This attr '
            'is set by \'auth\' method when called with \'groups\' Doc Arg for Controller Auth '
            'Sequence',
            list=[
                Attr.ID(
                    desc='\'_id\' of Group doc the session is authenticated against'
                )
            ],
        ),
        'host_add': Attr.IP(desc='IP of the host the user used to authenticate'),
        'user_agent': Attr.STR(
            desc='User-agent of the app the user used to authenticate'
        ),
        'expiry': Attr.DATETIME(
            desc='Python \'datetime\' ISO format of session expiry'
        ),
        'token_hash': Attr.STR(desc='Hashed system-generated session token'),
        'create_time': Attr.DATETIME(
            desc='Python \'datetime\' ISO format of the doc creation'
        ),
    },
    defaults={'groups': Default(value=[])},
    extns={'user': Extn(module='user', attrs=['*'], force=True)},
    funcs={
        'read': Func(
            permissions=[
                Perm(privilege='read', query_mod={'user': Var.SESSION('user')})
            ],
        ),
        'create': Func(permissions=[Perm(privilege='create')]),
        'update': Func(
            permissions=[
                Perm(
                    privilege='update',
                    query_mod={'user': Var.SESSION('user')},
                    doc_mod={'user': None},
                )
            ],
            query_attrs={'_id': Attr.ID()},
        ),
        'delete': Func(
            permissions=[
                Perm(privilege='__sys'),
            ],
            query_attrs={'_id': Attr.ID()},
        ),
        'auth': Func(
            permissions=[Perm(privilege='*')],
            call_args={'skip_status_check': Attr.BOOL()},
            doc_attrs=[],
            callable=_auth,
        ),
        'reauth': Func(
            permissions=[Perm(privilege='*')],
            query_attrs=[
                {
                    '_id': Attr.ID(),
                    'token': Attr.STR(),
                    'groups': Attr.LIST(list=[Attr.ID()]),
                },
                {'_id': Attr.ID(), 'token': Attr.STR()},
            ],
            callable=_reauth,
        ),
        'signout': Func(
            permissions=[Perm(privilege='*')],
            query_attrs={'_id': Attr.ID()},
            callable=_signout,
        ),
    },
)



from ._conn import create_conn
from ._create import create
from ._delete import delete
from ._drop import drop
from ._read import read
from ._update import update

__all__ = [
    'create_conn',
    'create',
    'delete',
    'drop',
    'read',
    'update',
]



'''Provides exceptions related to 'call' Utility'''


class InvalidCallEndpointException(Exception):
    '''Raised by 'call' Utility if value for arg 'endpoint' is invalid'''

    status = 400

    def __init__(self, *, endpoint):
        super().__init__(
            InvalidCallEndpointException.format_msg(endpoint=endpoint),
            {'endpoint': endpoint},
        )

    @staticmethod
    def format_msg(*, endpoint):
        '''Formats exception message'''

        return f'Invliad endpoint format: \'{endpoint}\''


class InvalidModuleException(Exception):
    '''Raised by 'call' Utility if endpoint points to non-existent Nawah Module'''

    status = 404

    def __init__(self, *, module_name):
        super().__init__(
            InvalidModuleException.format_msg(module_name=module_name),
            {'module_name': module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        '''Formats exception message'''

        return f'Nawah Module \'{module_name}\' is not defined'


class InvalidFuncException(Exception):
    '''Raised by 'call' Utility if endpoint points to non-existent Nawah Function'''

    status = 404

    def __init__(self, *, module_name, func_name):
        super().__init__(
            InvalidFuncException.format_msg(
                module_name=module_name, func_name=func_name
            ),
            {'module_name': module_name, 'func_name': func_name},
        )

    @staticmethod
    def format_msg(*, module_name, func_name):
        '''Formats exception message'''

        return f'Nawah Module \'{module_name}\' does not have Function \'{func_name}\' defined'


class NotPermittedException(Exception):
    '''Raised by 'check_permissions' Utility if failed to match Permissions Sets for current
    session user'''

    status = 403

    def __init__(self, *, module_name, func_name):
        super().__init__(
            NotPermittedException.format_msg(
                module_name=module_name, func_name=func_name
            ),
            {'module_name': module_name, 'func_name': func_name},
        )

    @staticmethod
    def format_msg(*, module_name, func_name):
        '''Formats exception message'''

        return f'Not permitted to access \'{module_name}\'.\'{func_name}\''



'''Nawah is rapid app development framework'''

import logging
import os

logger = logging.getLogger('nawah')
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s [%(levelname)s]  %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

logger.setLevel(logging.INFO)

__version__ = '0.0.0'
with open(
    os.path.join(os.path.dirname(__file__), 'version.txt'), encoding='UTF-8'
) as f:
    __version__ = f.read().strip()



'''Provides exceptions related to validation'''


class MissingAttrException(Exception):
    '''Raised by 'validate_doc' Utility if required 'attr' is missing from 'doc' '''

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(
            MissingAttrException.format_msg(attr_name=attr_name),
            {'attr_name': attr_name},
        )

    @staticmethod
    def format_msg(*, attr_name):
        '''Formats exception message'''

        return f'Missing attr \'{attr_name}\''


class InvalidAttrException(Exception):
    '''Raised by 'validate_attr' Utility if 'attr' has invalid value'''

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            InvalidAttrException.format_msg(
                attr_name=attr_name, attr_type=attr_type, val_type=val_type
            ),
            {'attr_name': attr_name, 'attr_type': attr_type, 'val_type': val_type},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_type, val_type):
        '''Formats exception message'''

        return f'Invalid attr \'{attr_name}\' of type \'{val_type}\' with required type \'{attr_type}\''



'''Provides 'delete_lock' Base Function callable'''

import logging
from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.config import Config
from nawah.enums import DeleteStrategy

from .exceptions import FailedDeleteLockException

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahEnv, Results

logger = logging.getLogger('nawah')


async def delete_lock(
    *, module_name: str, env: 'NawahEnv', query: 'Query'
) -> 'Results':
    '''Deletes locks for a module matching query \'query\'. If not, raises MethodException.'''

    lock_id = query['_id'][0]

    module = Config.modules[module_name]

    docs_results = results = await Data.read(
        env=env,
        collection_name=f'{module.collection}__lock',
        attrs={},
        query=query,
        skip_process=True,
    )
    results = await Data.delete(
        env=env,
        collection_name=f'{module.collection}__lock',
        docs=[doc['_id'] for doc in docs_results['docs']],
        strategy=DeleteStrategy.FORCE_SYS,
    )

    if results['count'] != 1:
        logger.error(
            'Failed to delete lock \'%s\' for \'%s\'. DELETE MANUALLY NOW',
            lock_id,
            module_name,
        )
        raise FailedDeleteLockException(module_name=module_name, lock_id=lock_id)

    return {
        'status': 200,
        'msg': f'Deleted {results["count"]} docs',
        'args': results,
    }



from ._cli import cli
from ._generate import generate_models_dart, generate_models_typescript
from ._serve import serve

__all__ = ['cli', 'generate_models_dart', 'generate_models_typescript', 'serve']



import aiohttp.web
from multidict import MultiDict

from nawah.classes import app_encoder
from nawah.config import Config


async def _not_found_handler(_):
    headers = MultiDict(
        [
            ('Server', 'Nawah'),
            ('Powered-By', 'Nawah, https://nawah.masaar.com'),
            ('Access-Control-Allow-Origin', '*'),
            ('Access-Control-Allow-Methods', 'GET,POST'),
            ('Access-Control-Allow-Headers', 'Content-Type'),
            ('Access-Control-Expose-Headers', 'Content-Disposition'),
        ]
    )
    return aiohttp.web.Response(
        status=404,
        headers=headers,
        body=app_encoder.encode({'status': 404, 'msg': '404 NOT FOUND'}),
    )


async def _not_allowed_handler(_):
    headers = MultiDict(
        [
            ('Server', 'Nawah'),
            ('Powered-By', 'Nawah, https://nawah.masaar.com'),
            ('Access-Control-Allow-Origin', '*'),
            ('Access-Control-Allow-Methods', '*'),
            ('Access-Control-Allow-Headers', 'Content-Type'),
            ('Access-Control-Expose-Headers', 'Content-Disposition'),
        ]
    )
    return aiohttp.web.Response(
        status=405,
        headers=headers,
        body=app_encoder.encode({'status': 405, 'msg': '404 NOT ALLOWED'}),
    )


async def _root_handler(_):
    headers = MultiDict(
        [
            ('Server', 'Nawah'),
            ('Powered-By', 'Nawah, https://nawah.masaar.com'),
            ('Access-Control-Allow-Origin', '*'),
            ('Access-Control-Allow-Methods', 'GET'),
            ('Access-Control-Allow-Headers', 'Content-Type'),
            ('Access-Control-Expose-Headers', 'Content-Disposition'),
        ]
    )
    return aiohttp.web.Response(
        status=200,
        headers=headers,
        body=app_encoder.encode(
            {
                'status': 200,
                'msg': f'Welcome to {Config.sys.name}!',
                'args': {'version': Config.sys.version, 'powered_by': 'Nawah'},
            }
        ),
    )



'''Provides 'retrieve_file' Base Function callable'''

import io
import logging
from typing import (TYPE_CHECKING, Any, MutableMapping, MutableSequence,
                    Optional, cast)

from PIL import Image

from nawah import data as Data
from nawah.classes import Query
from nawah.config import Config

from ._shared import FILE_ATTRS
from .exceptions import (FileNotFoundException, FileNotImageException,
                         InvalidDocException, UtilityModuleDataCallException)

if TYPE_CHECKING:
    from nawah.types import NawahEnv, Results

logger = logging.getLogger('nawah')


async def retrieve_file(
    *,
    module_name: str,
    env: 'NawahEnv',
    query: 'Query',
) -> 'Results':
    '''Finds file from doc in module, and returns it as part of results. If doc not found, raises
    'InvalidDocException'. If attr not found in doc, of file not matching file[s] in attr, raises
    'FileNotFoundException'. If a value for arg 'thumb_dims' is provided, generates a thumbnail, but
    raises 'FileNotImageException' if file not an image. If process to generate thumbnail fails, for
    any reasons, original file would be returned, error would be logged'''

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name='retrieve_file'
        )

    attr_name = query['attr'][0]
    filename = query['filename'][0]
    thumb_dims: Optional[MutableSequence[int]] = None
    if 'thumb' in query:
        thumb_dims = [int(dim) for dim in query['thumb'][0].split('x')]

    file_results = await Data.read(
        env=env,
        collection_name=f'{module.collection}__file',
        attrs=FILE_ATTRS,
        query=Query([{'_id': query['_id'][0]}]),
    )

    if not file_results['count']:
        raise InvalidDocException(doc_id=query['_id'][0])

    retrieved_file = file_results['docs'][0]['file']

    retrieved_file['content'] = _generate_thumb(
        thumb_dims=thumb_dims,
        module_name=module_name,
        doc_id=query['_id'][0],
        attr_name=attr_name,
        filename=filename,
        retrieved_file=retrieved_file,
    )

    return {
        'status': 200,
        'msg': 'File attached to response',
        'args': {
            'return': 'file',
            'docs': [
                {
                    '_id': query['_id'][0],
                    'name': retrieved_file['name'],
                    'type': retrieved_file['type'],
                    'lastModified': retrieved_file['lastModified'],
                    'size': retrieved_file['size'],
                    'content': retrieved_file['content'],
                }
            ],
        },
    }


def _retrieve_file_from_attr(*, doc_id, attr_name, attr, filename):
    '''Attempts to find file in attr. if failed, raises 'FileNotFoundException' '''

    retrieved_file = None

    if isinstance(attr, list):
        for item in attr:
            if item['name'] == filename:
                retrieved_file = item
                break
    elif isinstance(attr, dict):
        attr = cast(MutableMapping[str, Any], attr)
        if attr['name'] == filename:
            retrieved_file = attr

    if not retrieved_file:
        # [DOC] No filename match
        raise FileNotFoundException(
            doc_id=doc_id, attr_name=attr_name, file_name=filename
        )

    return retrieved_file


def _generate_thumb(
    *, thumb_dims, module_name, doc_id, attr_name, filename, retrieved_file
):
    '''Checks 'thumb_dims' value to generate thumbnail for image. if file is not image, raises
    'FileNotImageException'. If failed while generating thumbnail, logs error, and returns
    original file'''

    if not thumb_dims:
        return retrieved_file['content']

    if retrieved_file['type'].split('/')[0] != 'image':
        raise FileNotImageException(
            doc_id=doc_id,
            attr_name=attr_name,
            file_name=filename,
            file_type=retrieved_file['type'],
        )
    try:
        image = Image.open(io.BytesIO(retrieved_file['content']))
        image.thumbnail(thumb_dims)
        stream = io.BytesIO()
        image.save(stream, format=image.format)
        stream.seek(0)
        return stream.read()
    except Exception as e:  # pylint: disable=broad-except
        logger.error('Failed to generate thumbnail for:')
        logger.error('- Module     : \'%s\'', module_name)
        logger.error('- Doc \'_id\': \'%s\'', doc_id)
        logger.error('- Attr       : \'%s\'', attr_name)
        logger.error('- File       : \'%s\'', filename)
        logger.error('- Exception message: %s', e)

    return retrieved_file['content']



'''Provides 'AppJSONEncoder', 'CacheJSONEncoder' JSONEncoder implementations, instances'''

import datetime
import json

from bson import ObjectId


class AppJSONEncoder(json.JSONEncoder):
    '''JSONEncoder implementation that converts 'ObjectId', 'datetime' objects to str'''

    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)

        if isinstance(o, datetime.datetime):
            return o.isoformat()

        if isinstance(o, bytes):
            return True

        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return str(o)


app_encoder = AppJSONEncoder()



import logging

from nawah.config import Config

logger = logging.getLogger('nawah')


async def _close_session(session_id: str):
    # [TODO] Check necissity to implement session_lock for deleteing env object from sessions

    if session_id not in Config.sys.sessions:
        logger.debug(f'Skipped closing session #\'{session_id}\'')
        return

    logger.debug(
        f'Closing data connection for session #\'{Config.sys.sessions[session_id]["id"]}\''
    )
    Config.sys.sessions[session_id]['conn'].close()

    logger.debug('Done closing data connection')
    logger.debug(
        f'Websocket connection status: {not Config.sys.sessions[session_id]["ws"].closed}'
    )

    if not Config.sys.sessions[session_id]['ws'].closed:
        await Config.sys.sessions[session_id]['ws'].close()
    logger.debug(f'Websocket connection for session #\'{session_id}\' closed')

    del Config.sys.sessions[session_id]



'''Provides 'UserAttr' dataclass'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ._attr import Attr
    from ._default import Default
    from ._extn import Extn
    from ._func import Func


@dataclass(kw_only=True)
class UserAttr:
    '''UserAttr dataclass serves role of defining additional 'user' module attrs, as value for
    user_attrs Config Attr. When defined, Nawah adds attr to 'user' module attrs, and adds
    privilege to module matching attr. Function of name 'update_{attr}' would be added to module
    functions, defaulting to single Permission Set with attr as privilege, allowing any user with
    'user.{attr}' in privileges to update user doc value for the attr. If 'sanitise' is True,
    'user' module 'read' function callable would remove attr from results docs'''

    type: 'Attr'
    auth_attr: bool
    sanitise: bool
    default: Optional['Default'] = None
    extn: Optional['Extn'] = None
    func: Optional['Func'] = None



'''Provides Base Functions for Nawah modules'''

from ._create import create
from ._create_file import create_file
from ._delete import delete
from ._delete_file import delete_file
from ._delete_lock import delete_lock
from ._obtain_lock import obtain_lock
from ._read import read
from ._read_diff import read_diff
from ._retrieve_file import retrieve_file
from ._update import update
from ._update_file import update_file

__all__ = [
    'create',
    'create_file',
    'delete',
    'delete_lock',
    'delete_file',
    'obtain_lock',
    'read',
    'read_diff',
    'retrieve_file',
    'update',
    'update_file',
]



'''Provides 'obtain_lock' Base Function callable'''

import asyncio
import logging
from typing import TYPE_CHECKING

from pymongo.errors import DuplicateKeyError

import nawah.data as Data
from nawah.classes import Query
from nawah.config import Config
from nawah.enums import DeleteStrategy

from .exceptions import FailedDeleteLockException, FailedObtainLockException

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, Results

logger = logging.getLogger('nawah')


async def obtain_lock(
    *, module_name: str, env: 'NawahEnv', doc: 'NawahDoc'
) -> 'Results':
    '''Creates a lock for a module by creating a doc in lock collection and confirms
    it is the top lock doc. If not, deletes doc and raises 'FailedObtainLockException'.
    If deleting doc failed, raises 'FailedDeleteLockException' '''

    module = Config.modules[module_name]

    if 'total_attempts' not in doc:
        doc['total_attempts'] = doc['attempts']

    logger.debug('Attempting to obtain lock for \'%s\'', module_name)
    try:
        create_results = await Data.create(
            env=env,
            collection_name=f'{module.collection}__lock',
            doc=doc,
        )
        logger.debug(
            'Obtained intermediate lock \'%s\' for \'%s\'',
            create_results['docs'][0]['_id'],
            module_name,
        )
        read_results = await Data.read(
            env=env,
            collection_name=f'{module.collection}__lock',
            attrs={},
            query=Query(
                [
                    {'tags': {'$in': doc['tags']}} if doc['tags'] else {},
                    {'$sort': {'_id': 1}},
                ]
            ),
        )
        logger.debug('Attempting to read current lock for \'%s\'', module_name)

        if create_results['docs'][0]['_id'] == read_results['docs'][0]['_id']:
            logger.debug('Successfully obtained lock for \'%s\'', module_name)
            return {
                'status': 200,
                'msg': f'Obtained lock for \'{module_name}\'',
                'args': create_results,
            }
    except DuplicateKeyError:
        create_results = read_results = {'count': 0, 'docs': [{'_id': None}]}

    logger.warning(
        'Failed to obtain lock for \'%s\', with with created lock_id: \'%s\', but read \'%s\'',
        module_name,
        create_results['docs'][0]['_id'],
        read_results['docs'][0]['_id'],
    )

    if create_results['docs'][0]['_id']:
        logger.warning(
            'Deleting lock \'%s\', for \'%s\'',
            create_results['docs'][0]['_id'],
            module_name,
        )
        lock_delete_results = await Data.delete(
            env=env,
            collection_name=f'{module.collection}__lock',
            docs=[create_results['docs'][0]['_id']],
            strategy=DeleteStrategy.FORCE_SYS,
        )

        if lock_delete_results['count'] != 1:
            logger.error(
                'Failed to delete failed lock \'%s\' for \'%s\'. DELETE MANUALLY NOW',
                create_results['docs'][0]['_id'],
                module_name,
            )
            raise FailedDeleteLockException(
                module_name=module_name, lock_id=create_results['docs'][0]['_id']
            )

    if doc['attempts']:
        await asyncio.sleep(0.2)
        logger.warning('Reattempting to obtain lock for \'%s\'', module_name)
        doc['attempts'] -= 1
        return await obtain_lock(module_name=module_name, env=env, doc=doc)

    raise FailedObtainLockException(module_name=module_name)



'''Provides exceptions related to Attr Type'''


class InvalidAttrTypeException(Exception):
    '''Raied by 'validate_type' Utility if Attr Type object is not valid'''

    def __init__(self, *, attr_type):
        super().__init__(
            InvalidAttrTypeException.format_msg(attr_type=attr_type),
            {'attr_type': attr_type},
        )

    @staticmethod
    def format_msg(*, attr_type):
        '''Formats exception message'''

        return f'Unknown or invalid Attr Type \'{attr_type}\''


class InvalidAttrTypeRefException(Exception):
    '''Raied by 'validate_type' Utility if Attr Type 'TYPE' object referes to invalid type'''

    def __init__(self, *, attr_type):
        super().__init__(
            InvalidAttrTypeRefException.format_msg(attr_type=attr_type),
            {'attr_type': attr_type},
        )

    @staticmethod
    def format_msg(*, attr_type):
        '''Formats exception message'''

        return f'Attr Type \'{attr_type}\' refers to invalid \'type\''


class InvalidAttrTypeArgException(Exception):
    '''Raised by 'validate_arg' Utility if Attr Type Arg value is invalid'''

    def __init__(self, *, arg_name: str, arg_type, arg_val):
        super().__init__(
            InvalidAttrTypeArgException.format_msg(
                arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
            ),
            {'arg_name': arg_name, 'arg_type': arg_type, 'arg_val': arg_val},
        )

    @staticmethod
    def format_msg(*, arg_name, arg_type, arg_val):
        '''Formats exception message'''

        return f'Invalid Attr Type Arg for \'{arg_name}\' expecting type \'{arg_type}\' but got \'{arg_val}\'.'


class JSONPathNotFoundException(Exception):
    '''Raised by \'extract_attr\' Utility if provided didn't match any attr in 'scope' '''

    status = 400

    def __init__(self, *, scope, attr_path):
        super().__init__(
            JSONPathNotFoundException.format_msg(scope=scope, attr_path=attr_path),
            {'scope': scope, 'attr_path': attr_path},
        )

    @staticmethod
    def format_msg(*, scope, attr_path):
        '''Formats exception message'''

        return f'Failed to extract {attr_path} from scope: {scope}'



'''Provides 'session' Module Functions callables'''

import datetime
import secrets
from typing import TYPE_CHECKING, Optional

from passlib.hash import pbkdf2_sha512

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException
from nawah.utils import call

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, Results


async def _auth(
    env: 'NawahEnv', doc: 'NawahDoc', skip_status_check: Optional[None]
) -> 'Results':
    for attr in Config.modules['user'].unique_attrs:
        if attr in doc.keys():
            key = attr
            break
    user_query = [{key: doc[key], '$limit': 1}]
    if 'groups' in doc and doc['groups']:
        user_query.append(
            [{'groups': {'$in': doc['groups']}}, {'privileges': {'*': ['*']}}]
        )

    user_results = await call(
        'user/read',
        skip_events=[Event.PERM],
        env=env,
        query=user_query,
        args={'skip_sanitise_results': True},
    )
    if not user_results['args']['count'] or not pbkdf2_sha512.verify(
        doc['hash'],
        user_results['args']['docs'][0][f'{key}_hash'],
    ):
        raise FuncException(
            status=403,
            msg='Wrong auth credentials.',
            args={'code': 'INVALID_CREDS'},
        )

    user = user_results['args']['docs'][0]

    if skip_status_check is not True:
        if user['status'] in ['banned', 'deleted']:
            raise FuncException(
                status=403,
                msg=f'User is {user["status"]}.',
                args={'code': 'INVALID_USER'},
            )

        if user['status'] == 'disabled_password':
            raise FuncException(
                status=403,
                msg='User password is disabled.',
                args={'code': 'INVALID_USER'},
            )

    token = secrets.token_urlsafe(32)
    session = {
        'user': user['_id'],
        'groups': doc['groups'] if 'groups' in doc.keys() else [],
        'host_add': env['REMOTE_ADDR'],
        'user_agent': env['HTTP_USER_AGENT'],
        'expiry': (
            datetime.datetime.utcnow() + datetime.timedelta(days=30)
        ).isoformat(),
        'token_hash': pbkdf2_sha512.using(rounds=100000).hash(token),
    }

    results = await call(
        'session/create', skip_events=[Event.PERM], env=env, doc=session
    )
    if results['status'] != 200:
        return results

    session['_id'] = results['args']['docs'][0]['_id']
    session['user'] = user
    del session['token_hash']
    session['token'] = token
    results['args']['docs'][0] = session

    # read user privileges and return them
    user_results = await call(
        'user/read_privileges',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': user['_id']}],
    )
    if user_results['status'] != 200:
        return user_results
    results['args']['docs'][0]['user'] = user_results['args']['docs'][0]

    # Create CONN_AUTH Analytic doc
    # if Config.analytics_events['session_conn_auth']:
    #     analytic_doc = {
    #         'event': 'CONN_AUTH',
    #         'subevent': env['client_app'],
    #         'args': {
    #             'user': user_results['args']['docs'][0]['_id'],
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )
    # Create USER_AUTH Analytic doc
    # if Config.analytics_events['session_user_auth']:
    #     analytic_doc = {
    #         'event': 'USER_AUTH',
    #         'subevent': user_results['args']['docs'][0]['_id'],
    #         'args': {
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #             'client_app': env['client_app'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )

    return {
        'status': 200,
        'msg': 'You were successfully authed.',
        'args': {'session': results['args']['docs'][0]},
    }


async def _reauth(env: 'NawahEnv', query: 'Query') -> 'Results':
    if str(query['_id'][0]) == 'f00000000000000000000012':
        raise FuncException(
            status=400,
            msg='Reauth is not required for \'__ANON\' user.',
            args={'code': 'ANON_REAUTH'},
        )

    session_query = [{'_id': query['_id'][0]}]
    if query['groups'][0]:
        session_query.append({'groups': {'$in': query['groups'][0]}})
    results = await call(
        'session/read', skip_events=[Event.PERM], env=env, query=session_query
    )
    if not results['args']['count']:
        raise FuncException(
            status=403, msg='Session is invalid.', args={'code': 'INVALID_SESSION'}
        )

    if not pbkdf2_sha512.verify(
        query['token'][0], results['args']['docs'][0]['token_hash']
    ):
        raise FuncException(
            status=403,
            msg='Reauth token hash invalid.',
            args={'code': 'INVALID_REAUTH_HASH'},
        )

    del results['args']['docs'][0]['token_hash']
    results['args']['docs'][0]['token'] = query['token'][0]

    if results['args']['docs'][0]['expiry'] < datetime.datetime.utcnow().isoformat():
        results = await call(
            'session/delete',
            skip_events=[Event.PERM, Event.SOFT],
            env=env,
            query=[{'_id': env['session']['_id']}],
        )
        raise FuncException(
            status=403, msg='Session had expired.', args={'code': 'SESSION_EXPIRED'}
        )

    # update user's last_login timestamp
    await call(
        'user/update',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': results['args']['docs'][0]['user']}],
        doc={'login_time': datetime.datetime.utcnow().isoformat()},
    )
    await call(
        'session/update',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': results['args']['docs'][0]['_id']}],
        doc={
            'expiry': (
                datetime.datetime.utcnow() + datetime.timedelta(days=30)
            ).isoformat()
        },
    )
    # read user privileges and return them
    user_results = await call(
        'user/read_privileges',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': results['args']['docs'][0]['user']['_id']}],
    )
    results['args']['docs'][0]['user'] = user_results['args']['docs'][0]

    # Create CONN_AUTH Analytic doc
    # if Config.analytics_events['session_conn_reauth']:
    #     analytic_doc = {
    #         'event': 'CONN_REAUTH',
    #         'subevent': env['client_app'],
    #         'args': {
    #             'user': user_results['args']['docs'][0]['_id'],
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )
    # Create USER_AUTH Analytic doc
    # if Config.analytics_events['session_user_reauth']:
    #     analytic_doc = {
    #         'event': 'USER_REAUTH',
    #         'subevent': user_results['args']['docs'][0]['_id'],
    #         'args': {
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #             'client_app': env['client_app'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )

    return {
        'status': 200,
        'msg': 'You were successfully reauthed.',
        'args': {'session': results['args']['docs'][0]},
    }


async def _signout(env: 'NawahEnv', query: 'Query') -> 'Results':
    if str(query['_id'][0]) == 'f00000000000000000000012':
        raise FuncException(
            status=400,
            msg='Singout is not allowed for \'__ANON\' user.',
            args={'code': 'ANON_SIGNOUT'},
        )

    results = await call(
        'session/read',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': query['_id'][0]}],
    )

    if not results['args']['count']:
        raise FuncException(
            status=403, msg='Session is invalid.', args={'code': 'INVALID_SESSION'}
        )

    results = await call(
        'session/delete',
        skip_events=[Event.PERM],
        env=env,
        query=[{'_id': env['session']['_id']}],
    )

    # Create CONN_AUTH Analytic doc
    # if Config.analytics_events['session_conn_deauth']:
    #     analytic_doc = {
    #         'event': 'CONN_DEAUTH',
    #         'subevent': env['client_app'],
    #         'args': {
    #             'user': env['session']['user']['_id'],
    #             'session': env['session']['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )
    # Create USER_AUTH Analytic doc
    # if Config.analytics_events['session_user_deauth']:
    #     analytic_doc = {
    #         'event': 'USER_DEAUTH',
    #         'subevent': env['session']['user']['_id'],
    #         'args': {
    #             'session': env['session']['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #             'client_app': env['client_app'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )

    return {
        'status': 200,
        'msg': 'You are successfully signed-out.',
        'args': {'session': {'_id': 'f00000000000000000000012'}},
    }



'''Python package setup for Nawah Framework'''

from distutils.command.install import install

import setuptools

from nawah import __version__


class Version(install):
    '''Provides setup command to print Nawah Framework version'''

    def run(self):
        print(__version__)


class ApiLevel(install):
    '''Provides setup command to print Nawah Framework API Level'''

    def run(self):
        print('.'.join(__version__.split('.')[:2]))


with open('README.md', 'r', encoding='UTF-8') as f:
    long_description = f.read()

with open('./requirements.txt', 'r', encoding='UTF-8') as f:
    requirements = f.readlines()

with open('./dev_requirements.txt', 'r', encoding='UTF-8') as f:
    dev_requirements = f.readlines()

setuptools.setup(
    name='nawah',
    version=__version__,
    author='Mahmoud Abduljawad',
    author_email='mahmoud@masaar.com',
    description='Nawah framework--Rapid app development framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nawah-io/nawah_framework',
    package_data={
        'nawah': ['version.txt', 'py.typed'],
        'nawah.cli': ['py.typed'],
        'nawah.base': ['py.typed'],
        'nawah.classes': ['py.typed'],
        'nawah.exceptions': ['py.typed'],
        'nawah.types': ['py.typed'],
        'nawah.config': ['py.typed'],
        'nawah.data': ['py.typed'],
        'nawah.enums': ['py.typed'],
        'nawah.utils': ['py.typed'],
        'nawah.utils._app': ['py.typed'],
        'nawah.utils._validate': ['py.typed'],
        'nawah.testing': ['py.typed'],
        'nawah.packages': ['py.typed'],
        'nawah.packages.core': ['py.typed'],
        'nawah.packages.core.base': ['py.typed'],
        'nawah.packages.core.group': ['py.typed'],
        'nawah.packages.core.session': ['py.typed'],
        'nawah.packages.core.setting': ['py.typed'],
        'nawah.packages.core.user': ['py.typed'],
    },
    packages=[
        'nawah',
        'nawah.cli',
        'nawah.base',
        'nawah.classes',
        'nawah.exceptions',
        'nawah.types',
        'nawah.config',
        'nawah.data',
        'nawah.enums',
        'nawah.utils',
        'nawah.utils._app',
        'nawah.utils._validate',
        'nawah.testing',
        'nawah.packages',
        'nawah.packages.core',
        'nawah.packages.core',
        'nawah.packages.core.base',
        'nawah.packages.core.group',
        'nawah.packages.core.session',
        'nawah.packages.core.setting',
        'nawah.packages.core.user',
    ],
    project_urls={
        'Docs: Github': 'https://github.com/nawah-io/nawah_docs',
        'GitHub: issues': 'https://github.com/nawah-io/nawah_framework/issues',
        'GitHub: repo': 'https://github.com/nawah-io/nawah_framework',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: AsyncIO',
    ],
    python_requires='>=3.10.2',
    install_requires=requirements,
    extras_require={'dev': dev_requirements},
    cmdclass={
        'version': Version,
        'api_level': ApiLevel,
    },
    entry_points={
        'console_scripts': {
            'nawah = nawah.__main__:main',
        }
    },
    zip_safe=False,
)



'''Provides 'core' Nawah Package'''

from nawah.classes import Package

from .base import base
from .group import group
from .session import session
from .setting import setting
from .user import user

core = Package(
    name='core',
    api_level='2.0',
    version='2.0.0',
    modules=[base, user, group, session, setting],
)



'''Provides 'SysDoc' dataclass'''

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, cast

from bson import ObjectId

if TYPE_CHECKING:
    from nawah.types import NawahDoc

# [TODO] Create custom exceptions


@dataclass(kw_only=True)
class SysDoc:
    '''SysDoc dataclass serves role of defining items of 'docs' Config Attr which
    are docs created when an app is laucnhed, or as results of Nawah operations'''

    module: str
    key: Optional[str] = None
    skip_args: bool = False
    doc: Optional['NawahDoc'] = None

    @property
    def key_value(self) -> Any:
        '''Returns value for key 'key' from 'doc'. Raises ..., if 'doc' is not set'''

        if not self.doc:
            raise Exception('SysDoc instance was initialised with no \'doc\'.')
        self.key = cast(str, self.key)
        return self.doc[self.key]

    def __post_init__(self):
        if self.doc is not None:
            if not isinstance(self.doc, dict):
                raise Exception(
                    f'Argument \'doc\' is not a valid \'NawahDoc\'. Expecting type \'dict\' but got \'{type(self.doc)}\'.'
                )
            self.doc = cast('NawahDoc', self.doc)
            if not self.key:
                self.key = '_id'
            if self.key not in self.doc.keys():
                raise Exception(f'Attr \'{self.key}\' is not present on \'doc\'.')
            if self.key == '_id' and not isinstance(self.doc['_id'], ObjectId):
                raise Exception(
                    f'Invalid attr \'_id\' of type \'{type(self.doc["_id"])}\' with required type \'ID\''
                )
        else:
            if self.key or self.skip_args:
                raise Exception(
                    'Arguments \'attr, skip_args\' should only be used with \'doc\'.'
                )



"""
Author       : zhangxianbing
Date         : 2020-12-27 09:22:14
LastEditors  : zhangxianbing
LastEditTime : 2021-06-04 10:10:10
Description  : JSONPath
"""
__version__ = "1.0.5"
__author__ = "zhangxianbing"

import json
import logging
import os
import re
import sys
from collections import defaultdict
from typing import Union


def create_logger(name: str = None, level: Union[int, str] = logging.INFO):
    """Get or create a logger used for local debug."""

    formater = logging.Formatter(
        f"%(asctime)s-%(levelname)s-[{name}] %(message)s", datefmt="[%Y-%m-%d %H:%M:%S]"
    )

    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(formater)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger


logger = create_logger("jsonpath", os.getenv("PYLOGLEVEL", "INFO"))


class ExprSyntaxError(Exception):
    pass


class JSONPath:
    RESULT_TYPE = {
        "VALUE": "A list of specific values.",
        "PATH": "All path of specific values.",
    }

    # common patterns
    SEP = ";"
    REP_DOUBLEDOT = re.compile(r"\.\.")
    REP_DOT = re.compile(r"(?<!\.)\.(?!\.)")

    # save special patterns
    REP_GET_QUOTE = re.compile(r"['](.*?)[']")
    REP_PUT_QUOTE = re.compile(r"#Q(\d+)")
    REP_GET_BACKQUOTE = re.compile(r"[`](.*?)[`]")
    REP_PUT_BACKQUOTE = re.compile(r"#BQ(\d+)")
    REP_GET_BRACKET = re.compile(r"[\[](.*?)[\]]")
    REP_PUT_BRACKET = re.compile(r"#B(\d+)")
    REP_GET_PAREN = re.compile(r"[\(](.*?)[\)]")
    REP_PUT_PAREN = re.compile(r"#P(\d+)")

    # operators
    REP_SLICE_CONTENT = re.compile(r"^(-?\d*)?:(-?\d*)?(:-?\d*)?$")
    REP_SELECT_CONTENT = re.compile(r"^([\w.']+)(, ?[\w.']+)+$")
    REP_FILTER_CONTENT = re.compile(
        r"@\.(.*?)(?=<=|>=|==|!=|>|<| in| not| is)|len\(@\.(.*?)\)"
    )

    # annotations
    f: list
    segments: list
    lpath: int
    subx = defaultdict(list)
    result: list
    result_type: str
    eval_func: callable

    def __init__(self, expr: str):
        expr = self._parse_expr(expr)
        self.segments = expr.split(JSONPath.SEP)
        self.lpath = len(self.segments)
        logger.debug(f"segments  : {self.segments}")

        self.caller_globals = sys._getframe(1).f_globals

    def parse(self, obj, result_type="VALUE", eval_func=eval):
        if not isinstance(obj, (list, dict)):
            raise TypeError("obj must be a list or a dict.")

        if result_type not in JSONPath.RESULT_TYPE:
            raise ValueError(
                f"result_type must be one of {tuple(JSONPath.RESULT_TYPE.keys())}"
            )
        self.result_type = result_type
        self.eval_func = eval_func

        self.result = []
        self._trace(obj, 0, "$")

        return self.result

    def search(self, obj, result_type="VALUE"):
        return self.parse(obj, result_type)

    def _parse_expr(self, expr):
        logger.debug(f"before expr : {expr}")
        # pick up special patterns
        expr = JSONPath.REP_GET_QUOTE.sub(self._get_quote, expr)
        expr = JSONPath.REP_GET_BACKQUOTE.sub(self._get_backquote, expr)
        expr = JSONPath.REP_GET_BRACKET.sub(self._get_bracket, expr)
        expr = JSONPath.REP_GET_PAREN.sub(self._get_paren, expr)
        # split
        expr = JSONPath.REP_DOUBLEDOT.sub(f"{JSONPath.SEP}..{JSONPath.SEP}", expr)
        expr = JSONPath.REP_DOT.sub(JSONPath.SEP, expr)
        # put back
        expr = JSONPath.REP_PUT_PAREN.sub(self._put_paren, expr)
        expr = JSONPath.REP_PUT_BRACKET.sub(self._put_bracket, expr)
        expr = JSONPath.REP_PUT_BACKQUOTE.sub(self._put_backquote, expr)
        expr = JSONPath.REP_PUT_QUOTE.sub(self._put_quote, expr)
        if expr.startswith("$;"):
            expr = expr[2:]

        logger.debug(f"after expr  : {expr}")
        return expr

    # TODO abstract get and put procedures
    def _get_quote(self, m):
        n = len(self.subx["#Q"])
        self.subx["#Q"].append(m.group(1))
        return f"#Q{n}"

    def _put_quote(self, m):
        return self.subx["#Q"][int(m.group(1))]

    def _get_backquote(self, m):
        n = len(self.subx["#BQ"])
        self.subx["#BQ"].append(m.group(1))
        return f"`#BQ{n}`"

    def _put_backquote(self, m):
        return self.subx["#BQ"][int(m.group(1))]

    def _get_bracket(self, m):
        n = len(self.subx["#B"])
        self.subx["#B"].append(m.group(1))
        return f".#B{n}"

    def _put_bracket(self, m):
        return self.subx["#B"][int(m.group(1))]

    def _get_paren(self, m):
        n = len(self.subx["#P"])
        self.subx["#P"].append(m.group(1))
        return f"(#P{n})"

    def _put_paren(self, m):
        return self.subx["#P"][int(m.group(1))]

    @staticmethod
    def _gen_obj(m):
        ret = "__obj"
        for e in m.group(1).split("."):
            ret += '["%s"]' % e
        return ret

    @staticmethod
    def _traverse(f, obj, i: int, path: str, *args):
        if isinstance(obj, list):
            for idx, v in enumerate(obj):
                f(v, i, f"{path}{JSONPath.SEP}{idx}", *args)
        elif isinstance(obj, dict):
            for k, v in obj.items():
                f(v, i, f"{path}{JSONPath.SEP}{k}", *args)

    @staticmethod
    def _getattr(obj: dict, path: str, *, convert_number_str=False):
        r = obj
        for k in path.split("."):
            try:
                r = r.get(k)
            except (AttributeError, KeyError) as err:
                logger.error(err)
                return None
        if convert_number_str and isinstance(r, str):
            try:
                if r.isdigit():
                    return int(r)
                return float(r)
            except ValueError:
                pass
        return r

    @staticmethod
    def _sorter(obj, sortbys):
        for sortby in sortbys.split(",")[::-1]:
            if sortby.startswith("~"):
                obj.sort(
                    key=lambda t, k=sortby: JSONPath._getattr(
                        t[1], k[1:], convert_number_str=True
                    ),
                    reverse=True,
                )
            else:
                obj.sort(
                    key=lambda t, k=sortby: JSONPath._getattr(
                        t[1], k, convert_number_str=True
                    )
                )

    def _filter(self, obj, i: int, path: str, step: str):
        r = False
        try:
            r = self.eval_func(step, None, {"__obj": obj})
        except Exception as err:
            logger.error(err)
        if r:
            self._trace(obj, i, path)

    def _trace(self, obj, i: int, path):
        """Perform operation on object.

        Args:
            obj ([type]): current operating object
            i (int): current operation specified by index in self.segments
        """

        # store
        if i >= self.lpath:
            if self.result_type == "VALUE":
                self.result.append(obj)
            elif self.result_type == "PATH":
                self.result.append(path)
            logger.debug(f"path: {path} | value: {obj}")
            return

        step = self.segments[i]

        # wildcard
        if step == "*":
            self._traverse(self._trace, obj, i + 1, path)
            return

        # recursive descent
        if step == "..":
            self._trace(obj, i + 1, path)
            self._traverse(self._trace, obj, i, path)
            return

        # get value from list
        if isinstance(obj, list) and step.isdigit():
            ikey = int(step)
            if ikey < len(obj):
                self._trace(obj[ikey], i + 1, f"{path}{JSONPath.SEP}{step}")
            return

        # get value from dict
        if isinstance(obj, dict) and step in obj:
            self._trace(obj[step], i + 1, f"{path}{JSONPath.SEP}{step}")
            return

        # slice
        if isinstance(obj, list) and JSONPath.REP_SLICE_CONTENT.fullmatch(step):
            obj = list(enumerate(obj))
            vals = self.eval_func(f"obj[{step}]")
            for idx, v in vals:
                self._trace(v, i + 1, f"{path}{JSONPath.SEP}{idx}")
            return

        # select
        if isinstance(obj, dict) and JSONPath.REP_SELECT_CONTENT.fullmatch(step):
            for k in step.split(","):
                if k in obj:
                    self._trace(obj[k], i + 1, f"{path}{JSONPath.SEP}{k}")
            return

        # filter
        if step.startswith("?(") and step.endswith(")"):
            step = step[2:-1]
            step = JSONPath.REP_FILTER_CONTENT.sub(self._gen_obj, step)
            self._traverse(self._filter, obj, i + 1, path, step)
            return

        # sorter
        if step.startswith("/(") and step.endswith(")"):
            if isinstance(obj, list):
                obj = list(enumerate(obj))
                self._sorter(obj, step[2:-1])
                for idx, v in obj:
                    self._trace(v, i + 1, f"{path}{JSONPath.SEP}{idx}")
            elif isinstance(obj, dict):
                obj = list(obj.items())
                self._sorter(obj, step[2:-1])
                for k, v in obj:
                    self._trace(v, i + 1, f"{path}{JSONPath.SEP}{k}")
            else:
                raise ExprSyntaxError("sorter must acting on list or dict")
            return

        # field-extractor
        if step.startswith("(") and step.endswith(")"):
            if isinstance(obj, dict):
                obj_ = {}
                for k in step[1:-1].split(","):
                    obj_[k] = self._getattr(obj, k)
                self._trace(obj_, i + 1, path)
            else:
                raise ExprSyntaxError("field-extractor must acting on dict")

            return


def compile(expr):
    return JSONPath(expr)


# global cache
_jsonpath_cache = {}


def search(expr, data):
    global _jsonpath_cache
    if expr not in _jsonpath_cache:
        _jsonpath_cache[expr] = JSONPath(expr)
    return _jsonpath_cache[expr].parse(data)
